"""The main AQUA Reader class"""

from contextlib import contextmanager
from glob import glob
import os
import re
#import intake_esm
import intake_xarray
import xarray as xr
import pandas as pd
from metpy.units import units

from smmregrid import GridInspector

from aqua.core.util import load_multi_yaml, files_exist, to_list, find_vert_coord
from aqua.core.configurer import ConfigPath
from aqua.core.logger import log_configure, log_history
from aqua.core.exceptions import NoDataError, NoRegridError
from aqua.core.version import __version__ as aqua_version
from aqua.core.regridder import Regridder
from aqua.core.fldstat import FldStat
from aqua.core.timstat import TimStat
from aqua.core.fixer import Fixer
from aqua.core.data_model import DataModel, counter_reverse_coordinate
from aqua.core.histogram import histogram
import aqua.core.gsv

from .streaming import Streaming
from .trender import Trender

from .reader_utils import set_attrs

# set default options for xarray
xr.set_options(keep_attrs=True)

# set default data model
DATA_MODEL_DEFAULT = "aqua"

class Reader():
    """General reader for climate data."""

    instance = None  # Used to store the latest instance of the class

    def __init__(self, model=None, exp=None, source=None, catalog=None,
                 fix=True,
                 datamodel=None,
                 regrid=None, regrid_method=None,
                 areas=True,
                 streaming=False,
                 startdate=None, enddate=None,
                 rebuild=False, loglevel=None, nproc=4,
                 aggregation=None, chunks=None,
                 preproc=None, convention='eccodes',
                 engine='fdb',
                 **kwargs):
        """
        Initializes the Reader class, which uses the catalog
        `config/config.yaml` to identify the required data.

        Args:
            model (str): Model ID. Mandatory
            exp (str): Experiment ID. Mandatory.
            source (str): Source ID. Mandatory
            catalog (str, optional): Catalog where to search for the triplet.  Default to None will allow for autosearch in
                                     the installed catalogs.
            datamodel (str, optional): Data model to apply for coordinate transformations (e.g., 'aqua'). Defaults to 'aqua'.
            regrid (str, optional): Perform regridding to grid `regrid`, as defined in `config/regrid.yaml`. Defaults to None.
            regrid_method (str, optional): CDO Regridding regridding method. Read from grid configuration.
                                           If not specified anywhere, using "ycon".
            fix (bool, optional): Activate data fixing
            areas (bool, optional): Compute pixel areas if needed. Defaults to True.
            streaming (bool, optional): If to retrieve data in a streaming mode. Defaults to False.
            startdate (str, optional): The starting date for reading/streaming the data (e.g. '2020-02-25'). Defaults to None.
            enddate (str, optional): The final date for reading/streaming the data (e.g. '2020-03-25'). Defaults to None.
            rebuild (bool, optional): Force rebuilding of area and weight files. Defaults to False.
            loglevel (str, optional): Level of logging according to logging module.
                                      Defaults to log_level_default of loglevel().
            nproc (int, optional): Number of processes to use for weights generation. Defaults to 4.
            aggregation (str, optional): the streaming frequency in pandas style (1M, 7D etc. or 'monthly', 'daily' etc.)
                                         Defaults to None (using default from catalog, recommended).
            chunks (str or dict, optional): chunking to be used for data access.
                                            Defaults to None (using default from catalog, recommended).
                                            If it is a string time chunking is assumed.
                                            If it is a dictionary the keys 'time' and 'vertical' are looked for.
                                            Time chunking can be one of S (step), 10M, 15M, 30M, h, 1h, 3h, 6h, D, 5D, W, M, Y.
                                            Vertical chunking is expressed as the number of vertical levels to be used.
            preproc (function, optional): a function to be applied to the dataset when retrieved. Defaults to None.
            convention (str, optional): convention to be used for reading data. Defaults to 'eccodes'.
                                        (Only one supported so far)
            engine (str, optional): Engine to be used for GSV retrieval: 'polytope' or 'fdb'. Defaults to 'fdb'.

        Keyword Args: 
            zoom (int, optional): HEALPix grid zoom level (e.g. zoom=10 is h1024). Allows for multiple gridname definitions.
            realization (int, optional): The ensemble realization number.
            **kwargs: Additional arbitrary keyword arguments to be passed as additional parameters to the intake catalog entry.

        Returns:
            Reader: A `Reader` class object.
        """

        Reader.instance = self  # record the latest instance of the class (used for accessor)

        # define the internal logger
        self.loglevel = loglevel
        self.logger = log_configure(log_level=self.loglevel, log_name='Reader')

        self.exp = exp
        self.model = model
        self.source = source
        # these infos are used by the regridder to correct define areas/weights name
        reader_kwargs = {'model': model, 'exp': exp, 'source': source}
        self.regrid_method = regrid_method
        self.nproc = nproc
        self.vert_coord = None
        self.time_correction = False  # extra flag for correction data with cumulation time on monthly timescale
        self.aggregation = aggregation
        self.chunks = chunks

        # Preprocessing function
        self.preproc = preproc

        # init the areas and dimensions
        self.grid_area = None
        self.src_grid_area = None
        self.tgt_grid_area = None
        self.src_space_coord = None
        self.tgt_space_coord = None
        self.vert_coord = None

        if streaming:
            self.streamer = Streaming(startdate=startdate,
                                      enddate=enddate,
                                      aggregation=aggregation,
                                      loglevel=self.loglevel)
            # Export streaming methods TO DO: probably useless
            self.reset_stream = self.streamer.reset
            self.stream = self.streamer.stream
        self.streaming = streaming

        self.startdate = startdate
        self.enddate = enddate

        self.sample_data = None  # used to avoid multiple calls of retrieve_plain

        # define configuration file and paths
        configurer = ConfigPath(catalog=catalog, loglevel=loglevel)
        self.configdir = configurer.configdir
        self.machine = configurer.get_machine()
        self.config_file = configurer.config_file
        self.cat, self.catalog_file, self.machine_file = configurer.deliver_intake_catalog(
            catalog=catalog, model=model, exp=exp, source=source)
        self.fixer_folder, self.grids_folder = configurer.get_reader_filenames()

        # deduce catalog name
        self.catalog = self.cat.name

        # machine dependent catalog path
        machine_paths, intake_vars = configurer.get_machine_info()

        # load the catalog
        aqua.core.gsv.GSVSource.first_run = True  # Hack needed to avoid double checking of paths (which would not work if on another machine using polytope)
        self.expcat = self.cat(**intake_vars)[self.model][self.exp]  # the top-level experiment entry
        # We open before without kwargs to filter kwargs which are not in the parameters allowed by the intake catalog entry
        self.esmcat = self.expcat[self.source]()

        self.kwargs = self._filter_kwargs(kwargs, engine=engine, intake_vars=intake_vars)
        self.kwargs = self._format_realization_reader_kwargs(self.kwargs)
        self.logger.debug("Using filtered kwargs: %s", self.kwargs)
        self.esmcat = self.expcat[self.source](**self.kwargs)

        # Manual safety check for netcdf sources (see #943), we output a more meaningful error message
        if isinstance(self.esmcat, intake_xarray.netcdf.NetCDFSource):
            if not files_exist(self.esmcat.urlpath):
                raise NoDataError(f"No NetCDF files available for {self.model} {self.exp} {self.source}, please check the urlpath: {self.esmcat.urlpath}")  # noqa E501

        # extend the unit registry
        units_extra_definition()
        # Get fixes dictionary and find them
        self.fix = fix  # fix activation flag
        self.fixer_name = self.esmcat.metadata.get('fixer_name', None)
        self.convention = convention
        if self.convention is not None and self.convention != 'eccodes':
            raise ValueError(f"Convention {self.convention} not supported, only 'eccodes' is supported so far.")

        # case to disable automatic fix
        if self.fixer_name is False:
            self.logger.warning('A False flag is specified in fixer_name metadata, disabling fix!')
            self.fix = False

        # Initialize variable fixer
        if self.fix:
            self.fixes_dictionary = load_multi_yaml(self.fixer_folder, loglevel=self.loglevel)
            self.fixer = Fixer(fixer_name=self.fixer_name,
                               convention=self.convention,
                               fixes_dictionary=self.fixes_dictionary,
                               metadata=self.esmcat.metadata,
                               loglevel=self.loglevel)
        
        # if data model is not passed to Reader, try to get it from the catalog source metadata
        if datamodel is None:
            self.datamodel_name = self.esmcat.metadata.get('data_model', DATA_MODEL_DEFAULT) 
        else:
            self.datamodel_name = datamodel

        # if datamodel is False, disable data model application
        if not self.datamodel_name:
            self.datamodel = None
            self.logger.warning("Data model is not specified, many AQUA functionalities will not work properly!")
        else:
            self.datamodel = DataModel(name=self.datamodel_name, loglevel=self.loglevel)
    
            
        # define grid names
        self.src_grid_name = self.esmcat.metadata.get('source_grid_name')
        self.tgt_grid_name = regrid

        # init the regridder and the areas
        self._configure_regridder(machine_paths, regrid=regrid, areas=areas,
                                  rebuild=rebuild, reader_kwargs=reader_kwargs)
        
        # init the fldstat modules. if areas are not available, will issue a warning
        cell_area = self.src_grid_area.cell_area if areas else None
        self.src_fldstat = FldStat(
            cell_area, grid_name=self.src_grid_name,
            horizontal_dims=self.src_space_coord, loglevel=self.loglevel
            )
        self.tgt_fldstat = None
        if regrid:
            if not areas:
                self.logger.warning("Regridding requires info on areas. As areas can usually be generated with smmregrid, setting areas to 'True'")
                areas = True
            self.tgt_fldstat = FldStat(
                self.tgt_grid_area.cell_area, grid_name=self.tgt_grid_name,
                horizontal_dims=self.tgt_space_coord, loglevel=self.loglevel
                )
            
        self.trender = Trender(loglevel=self.loglevel)

    def _configure_regridder(self, machine_paths, regrid=False, areas=False,
                             rebuild=False, reader_kwargs=None):
        """
        Configure the regridder and generate areas and weights.

        Arguments:
            machine_paths (dict): The machine specific paths. Used to configure regridder file paths
            regrid (bool): If regrid is required. Defaults to False.
            areas (bool): If areas are required. Defaults to False.
            rebuild (bool): If weights and areas should be rebuilt. Defaults to False.
            reader_kwargs (dict): The reader kwargs.
        """

        # load and check the regrid
        if regrid or areas:

            # create the configuration dictionary
            cfg_regrid = load_multi_yaml(folder_path=self.grids_folder,
                                         definitions=machine_paths['paths'],
                                         loglevel=self.loglevel)
            cfg_regrid = {**machine_paths, **cfg_regrid}

            if self.src_grid_name is None:
                self.logger.info('Grid metadata is not defined. Trying to access the real data')
                data = self._retrieve_plain()
                self.regridder = Regridder(cfg_regrid, data=data, loglevel=self.loglevel)
            elif self.src_grid_name is False:
                self.logger.info('Grid metadata is False, regrid and areas disabled')
                regrid = False
                areas = False
                return
            else:
                self.logger.info('Grid metadata is %s', self.src_grid_name)
                self.regridder = Regridder(cfg_regrid, src_grid_name=self.src_grid_name, 
                                           loglevel=self.loglevel)

                if self.regridder.error:
                    self.logger.warning('Issues in the Regridder() init: trying with data')
                    data = self._retrieve_plain()
                    self.regridder = Regridder(cfg_regrid, src_grid_name=self.src_grid_name, 
                                               data=data, loglevel=self.loglevel)

            # export src space coord and vertical coord
            self.src_space_coord = self.regridder.src_horizontal_dims
            self.vert_coord = self.regridder.src_vertical_dim

            # TODO: it is likely there are other cases where we need to disable regrid.
            if not self.regridder.cdo:
                areas = False
                regrid = False

        if areas:
            # generate source areas and expose them in the reader
            self.src_grid_area = self.regridder.areas(rebuild=rebuild, reader_kwargs=reader_kwargs)
            # apply optional fixes to areas
            if self.fix:
                self.src_grid_area = self.fixer.fixerdatamodel.apply(self.src_grid_area)
            # Apply data model transformation to areas
            if self.datamodel:
                self.src_grid_area = self.datamodel.apply(self.src_grid_area)

        # configure regridder and generate weights
        if regrid:
            # generate weights and init the SMMregridder
            self.regridder.weights(
                rebuild=rebuild,
                tgt_grid_name=self.tgt_grid_name,
                regrid_method=self.regrid_method,
                reader_kwargs=reader_kwargs)

        # generate destination areas, expose them and the associated space coordinates
        if areas and regrid:
            self.tgt_grid_area = self.regridder.areas(tgt_grid_name=self.tgt_grid_name, rebuild=rebuild)
            # apply optional fixes to areas
            if self.fix:
                self.tgt_grid_area = self.fixer.fixerdatamodel.apply(self.tgt_grid_area)
            # Apply data model transformation to target areas
            if self.datamodel:
                self.tgt_grid_area = self.datamodel.apply(self.tgt_grid_area, flip_coords=False)
            # expose target horizontal dimensions
            self.tgt_space_coord = self.regridder.tgt_horizontal_dims

        # activate time statistics
        self.timemodule = TimStat(loglevel=self.loglevel)

    def retrieve(self, var=None, level=None,
                 startdate=None, enddate=None,
                 history=True, sample=False):
        """
        Perform a data retrieve.

        Arguments:
            var (str, list): the variable(s) to retrieve. Defaults to None. If None, all variables are retrieved.
            level (list, float, int): Levels to be read, overriding default in catalog source (only for FDB) .
            startdate (str): The starting date for reading/streaming the data (e.g. '2020-02-25'). Defaults to None.
            enddate (str): The final date for reading/streaming the data (e.g. '2020-03-25'). Defaults to None.
            history (bool): If you want to add to the metadata history information about retrieve. Defaults to True.
            sample (bool): read only one default variable (used only if var is not specified). Defaults to False.

        Returns:
            A xarray.Dataset containing the required data.
        """

        if not startdate:  # In case the streaming startdate is used also for FDB copy it
            startdate = self.startdate
        if not enddate:  # In case the streaming startdate is used also for FDB copy it
            enddate = self.enddate

        # get loadvar
        if var:
            if isinstance(var, str) or isinstance(var, int):
                var = str(var).split()  # conversion to list guarantees that a Dataset is produced
            self.logger.info("Retrieving variables: %s", var)
            # HACK: to be checked if can be done in a better way
            loadvar = self.fixer.get_fixer_varname(var) if self.fix else var
        else:
            # If we are retrieving from fdb we have to specify the var
            if isinstance(self.esmcat, aqua.core.gsv.intake_gsv.GSVSource):

                metadata = self.esmcat.metadata
                if metadata:
                    loadvar = metadata.get('variables')

                    if loadvar is None:
                        loadvar = [self.esmcat._request['param']]  # retrieve var from catalog

                    if not isinstance(loadvar, list):
                        loadvar = [loadvar]

                    if sample:
                        # self.logger.debug("FDB source sample reading, selecting only one variable")
                        loadvar = [loadvar[0]]

                    self.logger.debug("FDB source: loading variables as %s", loadvar)

            else:
                loadvar = None

        ffdb = False
        
        # If this is an ESM-intake catalog use first dictionary value,
        #if isinstance(self.esmcat, intake_esm.core.esm_datastore):
        #    data = self.reader_esm(self.esmcat, loadvar)
        # If this is an fdb entry
        if isinstance(self.esmcat, aqua.core.gsv.intake_gsv.GSVSource):
            data = self.reader_fdb(self.esmcat, loadvar, startdate, enddate,
                                   dask=True, level=level)
            ffdb = True  # These data have been read from fdb
        else:
            data = self.reader_intake(self.esmcat, var, loadvar)

        # if retrieve history is required (disable for retrieve_plain)
        if history:
            fkind = "FDB" if ffdb else "file from disk"
            data = log_history(data, f"Retrieved from {self.model}_{self.exp}_{self.source} using {fkind}")

        if not ffdb:  # FDB sources already have the index, already selected levels
            data = self._add_index(data)  # add helper index
            data = self._select_level(data, level=level)  # select levels (optional)

        # Apply variable fixes (units, names, attributes) and data model fixes
        if self.fix:
            self.logger.debug("Applying variable fixes")
            data = self.fixer.fixer(data, var)
            data = self.fixer.fixerdatamodel.apply(data)

        # Apply base data model transformation (always applied)
        if self.datamodel:
            self.logger.debug("Applying base data model: %s", self.datamodel_name)
            data = self.datamodel.apply(data)

        # log an error if some variables have no units
        if isinstance(data, xr.Dataset) and self.fix:
            for variable in data.data_vars:
                if not hasattr(data[variable], 'units'):
                    self.logger.warning('Variable %s has no units!', variable)

        if self.streaming:
            data = self.streamer.stream(data)
        else:
            if data is None or len(data.data_vars) == 0:
                self.logger.error("Retrieved empty dataset for var=%s. First, check its existence in the data catalog.", var)

            if (startdate or enddate) and not ffdb:  # do not select if data come from FDB (already done)
                data = data.sel(time=slice(startdate, enddate))

        if isinstance(data, xr.Dataset):
            data.aqua.set_default(self)  # This links the dataset accessor to this instance of the Reader class

        # Preprocessing function
        if self.preproc:
            data = self.preproc(data)

        # Add info metadata in each dataset
        info_metadata = {'AQUA_model': self.model, 'AQUA_exp': self.exp,
                         'AQUA_source': self.source, 'AQUA_catalog': self.catalog,
                         'AQUA_version': aqua_version}
        for kwarg in self.kwargs:
            info_metadata[f'AQUA_{kwarg}'] = str(self.kwargs[kwarg])

        data = set_attrs(data, info_metadata)

        return data

    def _add_index(self, data):

        """
        Add a helper idx_{dim3d} coordinate to the data to be used for level selection

        Arguments:
            data (xr.Dataset):  the input xarray.Dataset

        Returns:
            A xarray.Dataset containing the data with the idx_dim{3d} coordinate.
        """

        if self.vert_coord:
            for dim in to_list(self.vert_coord):
                if dim in data.coords:
                    idx = list(range(0, len(data.coords[dim])))
                    data = data.assign_coords(**{f"idx_{dim}": (dim, idx)})
        return data

    def _select_level(self, data, level=None):
        """
        Select levels if provided. It is based on self.vert_coord but it extends the feature 
        to atmospheric levels, so it should not be considered as the same vertical coordinate

        Arguments:
            data (xr.Dataset):  the input xarray.Dataset
            level (list, int):  levels to be selected. Defaults to None.

        Returns:
            A xarray.Dataset containing the data with the selected levels.
        """

        # return if no levels are selected
        if not level:
            return data

        # find the vertical coordinate, which can be the smmregrid one or 
        # any other with a dimension compatible (Pa, cm, etc)
        full_vert_coord = find_vert_coord(data) if not self.vert_coord else self.vert_coord

        # return if no vertical coordinate is found
        if not full_vert_coord:
            self.logger.error("Levels selected but no vertical coordinate found in data!")
            return data
        
        # ensure that level is a list
        level = to_list(level)

        # do the selection on the first vertical coordinate found
        if len(full_vert_coord) > 1:
            self.logger.warning(
                "Found more than one vertical coordinate, using the first one: %s", 
                full_vert_coord[0])
        data = data.sel(**{full_vert_coord[0]: level})
        data = log_history(data, f"Selecting levels {level} from vertical coordinate {full_vert_coord[0]}")
        return data

    def select_area(self, data, lon=None, lat=None, **kwargs):
        """
        Select a specific area from the dataset based on longitude and latitude ranges.

        Args:
            lon (list, optional): Longitude limits for the area selection.
            lat (list, optional): Latitude limits for the area selection.
            **kwargs: Additional keyword arguments to pass to the selection function. (See AreaSelection)
        """
        # We're keeping the fldstat call separate, however at the current stage there is
        # no difference in behavior between the src and tgt fldstat calls.
        if self._check_if_regridded(data):
            return self.tgt_fldstat.select_area(data, lon=lon, lat=lat, **kwargs)
        return self.src_fldstat.select_area(data, lon=lon, lat=lat, **kwargs)

    def set_default(self):
        """Sets this reader as the default for the accessor."""

        Reader.instance = self  # Refresh the latest reader instance used

    def regrid(self, data):
        """Call the regridder function returning container or iterator"""

        if self.tgt_grid_name is None:
            raise NoRegridError('regrid has not been initialized in the Reader, cannot perform any regrid.')
        
        data = counter_reverse_coordinate(data)

        out = self.regridder.regrid(data)

        # set regridded attribute to 1 for all vars
        out = set_attrs(out, {"AQUA_regridded": 1})
        return out
    
    # def trend(self, data, dim='time', degree=1, skipna=False):
    #     """
    #     Estimate the trend of an xarray object using polynomial fitting.

    #     Args:
    #         data (DataArray or Dataset): The input data.
    #         dim (str): Dimension to apply trend along. Defaults to 'time'.
    #         degree (int): Degree of the polynomial. Defaults to 1.
    #         skipna (bool): Whether to skip NaNs. Defaults to False.

    #     Returns:
    #         DataArray or Dataset: The trend component.
    #     """
    #     final = self.trender.trend(data, dim=dim, degree=degree, skipna=skipna)
    #     final.aqua.set_default(self)
    #     return final

    def detrend(self, data, dim='time', degree=1, skipna=False):    
        """
        Remove the trend from an xarray object using polynomial fitting.

        Args:
            data (DataArray or Dataset): The input data.
            dim (str): Dimension to apply detrend along. Defaults to 'time'.
            degree (int): Degree of the polynomial. Defaults to 1.
            skipna (bool): Whether to skip NaNs. Defaults to False.

        Returns:
            DataArray or Dataset: The detrended data.
        """
        final = self.trender.detrend(data, dim=dim, degree=degree, skipna=skipna)
        final.aqua.set_default(self)
        return final

    def _check_if_regridded(self, data):
        """
        Checks if a dataset or Datarray has been regridded.

        Arguments:
            data (xr.DataArray or xarray.DataDataset):  the input data
        Returns:
            A boolean value
        """

        if isinstance(data, xr.Dataset):
            att = list(data.data_vars.values())[0].attrs
        else:
            att = data.attrs

        return att.get("AQUA_regridded", False)

    # def _clean_spourious_coords(self, data, name=None):
    #     """
    #     Remove spurious coordinates from an xarray DataArray or Dataset.

    #     This function identifies and removes unnecessary coordinates that may
    #     be incorrectly associated with spatial coordinates, such as a time
    #     coordinate being linked to latitude or longitude.

    #     Parameters:
    #     ----------
    #     data : xarray.DataArray or xarray.Dataset
    #         The input data object from which spurious coordinates will be removed.

    #     name : str, optional
    #         An optional name or identifier for the data. This will be used in
    #         warning messages to indicate which dataset the issue pertains to.

    #     Returns:
    #     -------
    #     xarray.DataArray or xarray.Dataset
    #         The cleaned data object with spurious coordinates removed.
    #     """

    #     drop_coords = set()
    #     for coord in list(data.coords):
    #         if len(data[coord].coords)>1:
    #             drop_coords.update(koord for koord in data[coord].coords if koord != coord)
    #     if not drop_coords:
    #         return data
    #     self.logger.warning('Issue found in %s, removing %s coordinates',
    #                             name, list(drop_coords))
    #     return data.drop_vars(drop_coords)

    def _format_realization_reader_kwargs(self, kwargs: dict):
        """
        Reformats the realization string for the access to the reader
        If realization is in the format rXX and the intake type is int, it converts to int XX.
        """
        realization = kwargs.get('realization')
        if realization is None:
            return kwargs

        param_types = {p['name']: p['type'] for p in self.intake_user_parameters}
        realization_type = param_types.get('realization')

        if realization_type is None:
            self.logger.info("'realization' not in intake parameters %s — removing it.", list(param_types))
            kwargs.pop('realization', None)
            return kwargs

        # if type is string, return as is
        if realization_type == 'str':
            self.logger.debug('realization parameter is of type string, will use it is as is: %s', str(realization))
            kwargs['realization'] = str(realization)
            return kwargs

        # if it is in the rXX format and the type is int, convert to int
        if realization_type == 'int':
            if isinstance(realization, str) and realization.startswith('r') and realization[1:].isdigit():
                kwargs['realization'] = int(realization[1:])
                self.logger.info('realization parameter converted from rXXX format to int: %d', kwargs['realization'])
                return kwargs
            if isinstance(realization, int):
                return kwargs  # already an int

        raise ValueError(f"Realization {kwargs['realization']} format not recognized for type {realization_type}")

    @property
    def intake_user_parameters(self):
        """Lazy loader for intake user parameters to avoid expensive describe() calls."""
        if not hasattr(self, '_intake_user_parameters'):
            self._intake_user_parameters = self.esmcat.describe().get('user_parameters', {})
        return self._intake_user_parameters

    def _filter_kwargs(self, kwargs: dict={}, engine: str = 'fdb', intake_vars: dict = None):
        """
        Uses the esmcat.describe() to remove the intake_vars, then check in the parameters if the kwargs are present.
        Kwargs which are not present in the intake_vars will be removed.

        Args:
            kwargs (dict): The keyword arguments passed to the reader, which are intake parameters in the source.
            engine (str): The engine used for the GSV retrieval, default is 'fdb'.
            intake_vars (dict): Machine-specific intake variables to exclude from checks.

        Returns:
            A dictionary of kwargs filtered to only include parameters that are present in the intake_vars.
        """
        if intake_vars is None:
            intake_vars = {}

        filtered_kwargs = {}
        
        # Create a dictionary lookup of parameter definitions
        # This avoids repeated iteration and index lookups in the loop
        param_defs = {p['name']: p for p in self.intake_user_parameters}
        valid_params = set(param_defs.keys())

        if kwargs:
            # Filter kwargs that are valid parameters
            filtered_kwargs = {k: v for k, v in kwargs.items() if k in valid_params}
            
            # Find and log dropped keys efficiently using set difference
            dropped_keys = kwargs.keys() - valid_params
            for key in dropped_keys:
                self.logger.warning('kwarg %s is not in the intake parameters of the source, removing it', key)

        # Handle 'engine' for GSV/FDB sources
        if isinstance(self.esmcat, aqua.core.gsv.intake_gsv.GSVSource):
            if 'engine' not in filtered_kwargs:
                filtered_kwargs['engine'] = engine
                self.logger.debug('Adding engine=%s to the filtered kwargs', engine)

        # HACK: Keep chunking info if present as reader kwarg
        if self.chunks is not None:
            self.logger.warning('Keeping chunks=%s in the filtered kwargs', self.chunks)
            filtered_kwargs['chunks'] = self.chunks

        # Check for missing required parameters and apply defaults, with logging
        # We identify parameters that are valid but not present in either filtered_kwargs or intake_vars
        
        # params that are already covered by user kwargs or machine-specific intake_vars
        covered_params = set(filtered_kwargs) | set(intake_vars)
        
        # Identify missing parameters using set difference
        missing_params = valid_params - covered_params

        for param in missing_params:
            element = param_defs[param]
            default_val = element.get('default')
            
            # Log the default application
            self.logger.info('%s parameter is required but is missing, setting to default %s', param, default_val)
            
            allowed = element.get('allowed', None)
            if allowed is not None:
                self.logger.info('Available values for %s are: %s', param, allowed)
            
            filtered_kwargs[param] = default_val

        return filtered_kwargs

    def vertinterp(self, data, levels=None, vert_coord='plev', units=None,
                   method='linear'):
        """
        A basic vertical interpolation based on interp function
        of xarray within AQUA. Given an xarray object, will interpolate the
        vertical dimension along the vert_coord.
        If it is a Dataset, only variables with the required vertical
        coordinate will be interpolated.

        Args:
            data (DataArray, Dataset): your dataset
            levels (float, or list): The level you want to interpolate the vertical coordinate
            units (str, optional, ): The units of your vertical axis. Default 'Pa'
            vert_coord (str, optional): The name of the vertical coordinate. Default 'plev'
            method (str, optional): The type of interpolation method supported by interp()

        Return
            A DataArray or a Dataset with the new interpolated vertical dimension
        """

        if levels is None:
            raise KeyError('Levels for interpolation must be specified')

        # error if vert_coord is not there
        if vert_coord not in data.coords:
            raise KeyError(f'The vert_coord={vert_coord} is not in the data!')

        # if you not specified the units, guessing from the data
        if units is None:
            if hasattr(data[vert_coord], 'units'):
                self.logger.warning('Units of vert_coord=%s has not defined, reading from the data', vert_coord)
                units = data[vert_coord].units
            else:
                raise ValueError('Original dataset has not unit on the vertical axis, failing!')

        if isinstance(data, xr.DataArray):
            final = self._vertinterp(data=data, levels=levels, units=units,
                                     vert_coord=vert_coord, method=method)

        elif isinstance(data, xr.Dataset):
            selected_vars = [da for da in data.data_vars if vert_coord in data[da].coords]
            final = data[selected_vars].map(self._vertinterp, keep_attrs=True,
                                            levels=levels, units=units,
                                            vert_coord=vert_coord, method=method)
        else:
            raise ValueError('This is not an xarray object!')

        final = log_history(
            final, f"Interpolated from original levels {data[vert_coord].values} {data[vert_coord].units} to level {levels} using {method} method.") # noqa E501

        final.aqua.set_default(self)  # This links the dataset accessor to this instance of the Reader class

        return final

    def _vertinterp(self, data, levels=None, units='Pa', vert_coord='plev', method='linear'):

        # verify units are good
        if data[vert_coord].units != units:
            self.logger.warning('Converting vert_coord units to interpolate from %s to %s',
                                data[vert_coord].units, units)
            data = data.metpy.convert_coordinate_units(vert_coord, units)

        # very simple interpolation
        final = data.interp({vert_coord: levels}, method=method)

        return final

    def reader_esm(self, esmcat, var):
        """
        Read intake-esm entry. Returns a dataset.

        Args:
            esmcat (intake_esm.core.esm_datastore): The intake-esm catalog datastore to read from.
            var (str or list): Variable(s) to retrieve. If None, uses the query from catalog metadata.

        Returns:
            xarray.Dataset: The dataset retrieved from the intake-esm catalog.
        """
        xarray_open_kwargs = esmcat.metadata.get('xarray_open_kwargs', 
                                         esmcat.metadata.get('cdf_kwargs', {"chunks": {"time": 1}}))
        query = esmcat.metadata['query']
        if var:
            query_var = esmcat.metadata.get('query_var', 'short_name')
            # Convert to list if not already
            query[query_var] = var.split() if isinstance(var, str) else var
        subcat = esmcat.search(**query)
        data = subcat.to_dataset_dict(xarray_open_kwargs=xarray_open_kwargs,
                                      # zarr_kwargs=dict(consolidated=True),
                                      # decode_times=True,
                                      # use_cftime=True)
                                      progressbar=False
                                      )
        return list(data.values())[0]

    def reader_fdb(self, esmcat, var, startdate, enddate, dask=False, level=None):
        """
        Read fdb data. Returns a dask array.

        Args:
            esmcat (intake catalog): the intake catalog to read
            var (str, int or list): the variable(s) to read
            startdate (str): a starting date and time in the format YYYYMMDD:HHTT
            enddate (str): an ending date and time in the format YYYYMMDD:HHTT
            dask (bool): return directly a dask array
            level (list, float, int): level to be read, overriding default in catalog

        Returns:
            An xarray.Dataset 
        """
        # Var can be a list or a single one of these cases:
        # - an int, in which case it is a paramid
        # - a str, in which case it is a short_name that needs to be matched with the paramid
        # - a list (in this case I may have a list of lists) if fix=True and the original variable
        #   found a match in the source field of the fixer dictionary
        request = esmcat._request
        var = to_list(var)
        var_match = []

        fdb_var = esmcat.metadata.get('variables', None)
        # This is a fallback for the case in which no 'variables' metadata is defined
        # It is a backward compatibility feature and it may be removed in the future
        # We need to access with describe because the 'param' element is not a class
        # attribute. I would not add it since it is a deprecated feature
        if fdb_var is None:
            self.logger.warning("No 'variables' metadata defined in the catalog, this is deprecated!")
            fdb_var = esmcat.describe()["args"]["request"]["param"]
            fdb_var = to_list(fdb_var)

        # We avoid the following loop if the user didn't specify any variable
        # We make sure this is the case by checking that var is the same as fdb_var
        # If we need to loop, two cases may arise:
        # 1. fix=True: if elem is a paramid we try to match it with the list on fdb_var
        #              if is a str we scan in the fixer dictionary if there is a match
        #              and we use the paramid listed in the source block to match with fdb_var
        #              As a final fallback, if the scan fails, we use the initial str as a match
        #              letting eccodes itself to find the paramid (this may lead to errors)
        # 2. fix=False: we just scan the list of variables requested by the user.
        #               For paramids we do as case 1, while for str we just do as in the fallback
        #               option defined in case 1
        # We're trying to set the if/else by int vs str and then eventually by the fix option
        # We store the fixer_dict once for all for semplicity of the if case.
        if self.fix is True:
            fixer_dict = self.fixer.fixes.get('vars', {})
            if fixer_dict == {}:
                self.logger.debug("No 'vars' block in the fixer, guessing variable names base on ecCodes")
        if var != fdb_var:
            for element in var:
                # We catch also the case where we ask for var='137' but we know that is a paramid
                if isinstance(element, int) or (isinstance(element, str) and element.isdigit()):
                    element = int(element) if isinstance(element, str) else element
                    element = to_list(element)
                    match = list(set(fdb_var) & set(element))
                    if match and len(match) == 1:
                        var_match.append(match[0])
                    elif match and len(match) > 1:
                        self.logger.warning('Multiple matches found for %s, using the first one', element)
                        var_match.append(match[0])
                    else:
                        self.logger.warning('No match found for %s, skipping it', element)
                elif isinstance(element, str):
                    if self.fix is True:
                        if element in fixer_dict:
                            src_element = fixer_dict[element].get('source', None)
                            derived_element = fixer_dict[element].get('derived', None)
                            if derived_element is not None or src_element is None:  # We let eccodes to find the paramid
                                var_match.append(derived_element)
                            else:  # src_element is not None and it is not a derived variable
                                match = list(set(fdb_var) & set(src_element))
                                if match and len(match) == 1:
                                    var_match.append(match[0])
                                elif match and len(match) > 1:
                                    self.logger.warning('Multiple paramids found for %s: %s, using: %s',
                                                        element, match, match[0])
                                    var_match.append(match[0])
                                else:
                                    self.logger.warning('No match found for %s, using eccodes to find the paramid',
                                                        element)
                                    var_match.append(element)
                    else:
                        var_match.append(element)
                elif isinstance(element, list):
                    if self.fix is False:
                        raise ValueError(f"Var {element} is a list and fix is False, this is not allowed")
                    match = list(set(fdb_var) & set(element))
                    if match and len(match) == 1:
                        var_match.append(match[0])
                    elif match and len(match) > 1:
                        self.logger.warning('Multiple matches found for %s, using the first one', element)
                        var_match.append(match[0])
                    else:
                        self.logger.error('No match found for %s, skipping it', element)
                else:  # Something weird is happening, we may want to have a raise instead
                    self.logger.error("Element %s is not a valid type, skipping it", element)
        else:  # There is no need to scan the list of variables, total match
            var_match = var

        if var_match == []:
            self.logger.error("No match found for the variables you are asking for!")
            self.logger.error("Please be sure the metadata 'variables' is defined in the catalog")
            var_match = var
        else:
            self.logger.debug("Found variables: %s", var_match)

        var = var_match
        self.logger.debug("Requesting variables: %s", var)

        if level and not isinstance(level, list):
            level = [level]

        # for streaming emulator
        if self.aggregation and not dask:
            chunks = self.aggregation
        else:
            chunks = self.chunks

        if isinstance(chunks, dict):
            if self.aggregation and not chunks.get('time'):
                chunks['time'] = self.aggregation
            if self.streaming and not self.aggregation:
                self.logger.warning(
                    "Aggregation is not set, using default time resolution for streaming. If you are asking for a longer chunks['time'] for GSV access, please set a suitable aggregation value") # noqa E501

        if dask:
            if chunks:  # if the chunking or aggregation option is specified override that from the catalog
                data = esmcat(request=request, startdate=startdate, enddate=enddate, var=var, level=level,
                              chunks=chunks, logging=True, loglevel=self.loglevel).to_dask()
            else:
                data = esmcat(request=request, startdate=startdate, enddate=enddate, var=var, level=level,
                              logging=True, loglevel=self.loglevel).to_dask()
        else:
            if chunks:
                data = esmcat(request=request, startdate=startdate, enddate=enddate, var=var, level=level,
                              chunks=chunks, logging=True, loglevel=self.loglevel).read_chunked()
            else:
                data = esmcat(request=request, startdate=startdate, enddate=enddate, var=var, level=level,
                              logging=True, loglevel=self.loglevel).read_chunked()

        return data
    
    def _filter_netcdf_files(self, esmcat, filter_key="year"):

        """
        Filter the esmcat to include only netcdf files based on specfici filter_key
        Args:
            esmcat (intake.catalog.Catalog): your catalog
            filter_key (str): type of filter to apply (default is "year")

        Returns:
            intake.catalog.Catalog: filtered catalog
        """

        # list available files in folder
        files = sorted([f for x in esmcat.urlpath for f in glob(x)])
        self.logger.debug("Total files before filtering: %s", len(files))

        # this will consider only files that have "year" in their filename 
        # within the startdate and enddate range
        if filter_key == "year": 
            if not (self.startdate and self.enddate):
                return esmcat
            keys = list(range(pd.Timestamp(self.startdate).year,
                               pd.Timestamp(self.enddate).year + 1))
            # create regex pattern for each year: only yyyy will be detected
            pattern = [re.compile(rf'(?<!\d){yr}(?!\d)') for yr in keys]

        else:
            raise ValueError(f"Filter type {filter_key} not recognized.")

        # replace the urlpath with the filtered one searching the regex
        esmcat.urlpath = [f for f in files if any(p.search(os.path.basename(f)) for p in pattern)]

        if len(esmcat.urlpath) == 0:
            raise NoDataError("No files found after filtering the catalog!")
        
        self.logger.debug("Selected: %s files from %s to %s",
                          len(esmcat.urlpath), esmcat.urlpath[0], esmcat.urlpath[-1])

        return esmcat

    def reader_intake(self, esmcat, var, loadvar, keep="first"):
        """
        Read regular intake entry. Returns dataset.

        Args:
            esmcat (intake.catalog.Catalog): your catalog
            var (list or str): Variable to load
            loadvar (list of str): List of variables to load
            keep (str, optional): which duplicate entry to keep ("first" (default), "last" or None)

        Returns:
            Dataset
        """

        # use filter_key metadata to filter netcdf files if present
        # speed up for catalogs with many small files
        if 'filter_key' in esmcat.metadata and isinstance(self.esmcat, intake_xarray.netcdf.NetCDFSource):
            self.logger.info("Filtering netcdf files in the catalog based on %s", esmcat.metadata.get('filter_key'))
            esmcat = self._filter_netcdf_files(esmcat, filter_key=esmcat.metadata['filter_key'])
    
        # The coder introduces the possibility to specify a time decoder for the time axis
        if 'time_coder' in esmcat.metadata:
            self.logger.info('Using custom pandas/xarray time coder: %s', esmcat.metadata['time_coder'])
            coder = xr.coders.CFDatetimeCoder(time_unit=esmcat.metadata['time_coder'])
            esmcat.xarray_kwargs.update({'decode_times': coder})

        data = esmcat.to_dask()

        if loadvar:
            loadvar = to_list(loadvar)
            loadvar_match = []
            for element in loadvar:
                # Having to do a list comprehension we want to be sure that the element is a list
                element = to_list(element)
                match = list(set(data.data_vars) & set(element))

                if match:
                    loadvar_match.append(match[0])
                else:
                    self.logger.warning('No match found for %s', element)
            loadvar = loadvar_match

            if all(element in data.data_vars for element in loadvar):
                data = data[loadvar]
            else:
                try:
                    data = data[var]
                    self.logger.warning("You are asking for var %s but the fixes definition requires %s, which is not there.",
                                        var, loadvar)
                    self.logger.warning("Retrieving %s, but it would be safer to run with fix=False or to correct the fixes",
                                        var)
                except Exception as e:
                    raise KeyError("You are asking for variables which we cannot find in the catalog!") from e

        # check for duplicates
        if 'time' in data.coords:
            len0 = len(data.time)
            data = data.drop_duplicates(dim='time', keep=keep)
            if len(data.time) != len0:
                self.logger.warning("Duplicate entries found along the time axis, keeping the %s one.", keep)

        return data

    @contextmanager
    def _temporary_attrs(self, **kwargs):
        """Temporarily override Reader attributes, restoring them afterward."""
        original_values = {key: getattr(self, key) for key in kwargs}
        try:
            for key, value in kwargs.items():
                setattr(self, key, value)
            yield
        finally:
            for key, value in original_values.items():
                setattr(self, key, value)

    def _retrieve_plain(self, *args, **kwargs):
        """
        Retrieve data without any additional processing.
        Making use of GridInspector, provide a sample data which has minimum
        size by subselecting along variables and time dimensions.
        Uses Reader's startdate/enddate if set.

        Args:
            *args: arguments to be passed to retrieve
            **kwargs: keyword arguments to be passed to retrieve

        Returns:
            A xarray.Dataset containing the required miminal sample data.
        """
        if self.sample_data is not None:
            self.logger.debug('Sample data already availabe, avoid _retrieve_plain()')
            return self.sample_data

        # Temporarily disable unwanted settings
        with self._temporary_attrs(aggregation=None, chunks=None, 
                                   fix=False, streaming=False,
                                   preproc=None):
            self.logger.debug('Getting sample data through _retrieve_plain()...')
            data = self.retrieve(history=False, *args, **kwargs)

        self.sample_data = self._grid_inspector(data)
        return self.sample_data

    def _grid_inspector(self, data):
        """
        Use smmregrid GridInspector to get minimal sample data

        Args:
            data (xarray.Dataset): input data

        Returns:
            A xarray.Dataset containing the required miminal sample data.
        """

        # get gridtypes from smrregird
        gridinspect = GridInspector(data, loglevel=self.loglevel)
        gridtypes = gridinspect.get_gridtype()

        # get info on time dimensions and variables
        minimal_variables = gridinspect.get_gridtype_attr(gridtypes, 'variables')
        minimal_time = gridinspect.get_gridtype_attr(gridtypes, 'time_dims')

        if minimal_variables:
            self.logger.debug('Variables found: %s', minimal_variables)
            data = data[minimal_variables]
        if minimal_time:
            self.logger.debug('Time dimensions found: %s', minimal_time)
            data = data.isel({t: 0 for t in minimal_time})
        return data

    def fldstat(self, data, stat, lon_limits=None, lat_limits=None, dims=None, 
                region=None, region_sel=None, mask_kwargs={}, **kwargs):
        """
        Field statistic wrapper which is calling the fldstat module from FldStat class. 
        This method is exposing and providing field functions as Reader class 
        methods through the wrapper accessors.

        Args:
            data (xr.DataArray or xarray.Dataset):  the input data
            stat (str):  the statistical function to be applied
            lon_limits (list, optional):  the longitude limits of the subset
            lat_limits (list, optional):  the latitude limits of the subset
            dims (list, optional):  the dimensions to average over
            region (regionmask.Regions, optional): A regionmask Regions object defining a class regions.
            region_sel (str, int or list, optional): The region(s) to select by name or number from the region object.
            mask_kwargs (dict, optional): Additional keyword arguments passed to region.mask().
            **kwargs: additional arguments passed to fldstat
        """
        # Handle regridding logic - use appropriate fldstat module
        if self._check_if_regridded(data):
            data = self.tgt_fldstat.fldstat(
                data, stat=stat,
                lon_limits=lon_limits, lat_limits=lat_limits,
                region=region, region_sel=region_sel, mask_kwargs=mask_kwargs,
                dims=dims, **kwargs)
        else:
            data = self.src_fldstat.fldstat(
                data, stat=stat,
                lon_limits=lon_limits, lat_limits=lat_limits,
                region=region, region_sel=region_sel, mask_kwargs=mask_kwargs,
                dims=dims, **kwargs)

        data.aqua.set_default(self)
        return data

    # Field stats wrapper. If regridded, uses the target grid areas.
    def fldmean(self, data, **kwargs):
        """
        Field mean wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='mean', **kwargs)

    def fldmax(self, data, **kwargs):
        """
        Field max wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='max', **kwargs)

    def fldmin(self, data, **kwargs):
        """
        Field min wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='min', **kwargs)

    def fldstd(self, data, **kwargs):
        """
        Field standard deviation wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='std', **kwargs)

    def fldsum(self, data, **kwargs):
        """
        Field sum wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='sum', **kwargs)

    def fldintg(self, data, **kwargs):
        """
        Field integral wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='integral', **kwargs)

    def fldarea(self, data, **kwargs):
        """
        Field area wrapper which is calling the fldstat module.
        """
        return self.fldstat(data, stat='areasum', **kwargs)

    def timstat(self, data, stat, freq=None, exclude_incomplete=False,
                time_bounds=False, center_time=False, **kwargs):
        """
        Time statistic wrapper which is calling the timstat module from TimStat class. 
        This method is exposing and providing time functions as Reader class 
        methods through the wrapper accessors.

        Args:
            data (xr.DataArray or xarray.Dataset):  the input data
            stat (str):  the statistical function to be applied
            freq (str):  the frequency of the time average
            exclude_incomplete (bool):  exclude incomplete time averages
            time_bounds (bool):  produce time bounds after averaging
            center_time (bool):  center time for averaging
            kwargs:  additional arguments to be passed to the statistical function
        """
        data = self.timemodule.timstat(
            data, stat=stat, freq=freq,
            exclude_incomplete=exclude_incomplete,
            time_bounds=time_bounds,
            center_time=center_time, **kwargs)
        data.aqua.set_default(self) #accessor linking
        return data
    
    def timmean(self, data, **kwargs):
        """
        Time mean wrapper which is calling the timstat module.
        """
        return self.timstat(data, stat='mean', **kwargs)

    def timmax(self, data, **kwargs):
        """
        Time max wrapper which is calling the timstat module.
        """
        return self.timstat(data, stat='max', **kwargs)
    
    def timmin(self, data, **kwargs):
       """
       Time min wrapper which is calling the timstat module.
       """
       return self.timstat(data, stat='min', **kwargs)
    
    def timstd(self, data, **kwargs):
       """
       Time standard deviation wrapper which is calling the timstat module.
       """
       return self.timstat(data, stat='std', **kwargs)
    
    def timsum(self, data, **kwargs):
        """
        Time sum wrapper which is calling the timstat module.
        """
        return self.timstat(data, stat='sum', **kwargs)

    def timhist(self, data, **kwargs):
        """
        Wrapper for the histogram function, with added timstat functionality.
        It accepts arguments of timstat to resample in time before computing the histogram.
        """
        return self.timstat(data, stat=histogram, **kwargs)

    def histogram(self, data, **kwargs):
        """
        Wrapper for the histogram function
        """
        return histogram(data, **kwargs)

def units_extra_definition():
    """Add units to the pint registry"""

    # special units definition
    # needed to work with metpy 1.4.0 see
    # https://github.com/Unidata/MetPy/issues/2884
    units._on_redefinition = 'ignore'
    units.define('fraction = [] = Fraction = frac')
    units.define('psu = 1e-3 frac')
    units.define('PSU = 1e-3 frac')
    units.define('Sv = 1e+6 m^3/s')
    units.define("North = degrees_north = degreesN = degN")
    units.define("East = degrees_east = degreesE = degE")



"""Test for cli_checker command line interface"""

import os
import sys
import subprocess
import pytest
from aqua import __path__ as aqua_pkg_path
from aqua.core.analysis.cli_checker import parse_arguments

pytestmark = pytest.mark.aqua


def test_cli_checker_parse_arguments():
    """Test that parse_arguments correctly parses cli_checker specific arguments and defaults."""
    # Test with all flags provided
    args = parse_arguments([
        '--model', 'IFS',
        '--exp', 'test-tco79',
        '--source', 'short',
        '--yaml', '/tmp/test',
        '--no-read',
        '--no-rebuild',
        '--realization', 'r1',
        '--regrid', 'r200'
    ])
    
    assert args.model == 'IFS'
    assert args.exp == 'test-tco79'
    assert args.source == 'short'
    assert args.yaml == '/tmp/test'
    assert args.read is False  # --no-read sets read=False
    assert args.rebuild is False  # --no-rebuild sets rebuild=False
    assert args.realization == 'r1'
    assert args.regrid == 'r200'
    
    # Test defaults when flags are not provided
    args_defaults = parse_arguments([
        '--model', 'IFS',
        '--exp', 'test-tco79',
        '--source', 'short'
    ])
    
    assert args_defaults.yaml is None
    assert args_defaults.read is True  # Default is True (read data)
    assert args_defaults.rebuild is True  # Default is True (rebuild areas)
    assert args_defaults.realization is None


@pytest.mark.slow
def test_cli_checker_valid_entry(tmp_path):
    """Test cli_checker with a valid catalog entry.
    
    This test requires AQUA to be installed with the 'ci' catalog.
    Note: This test may be skipped if AQUA is not fully installed.
    """
    
    checker_script = os.path.join(aqua_pkg_path[0], "core", "analysis", "cli_checker.py")
    yaml_dir = tmp_path / "yaml_output"
    yaml_dir.mkdir()
    
    # Test with a valid entry from the 'ci' catalog, using --no-read to speed up
    result = subprocess.run(
        [
            sys.executable, checker_script,
            '--catalog', 'ci',
            '--model', 'IFS',
            '--exp', 'test-tco79',
            '--source', 'short',
            '--yaml', str(yaml_dir),
            '--no-read',  # Don't read data, just check catalog and create yaml
            '--no-rebuild',  # Don't rebuild areas
            '--loglevel', 'WARNING'
        ],
        capture_output=True,
        text=True,
        timeout=60
    )
    
    # Should succeed if AQUA is properly installed
    if result.returncode == 0:
        # Check that experiment.yaml was created
        yaml_file = yaml_dir / "experiment.yaml"
        assert yaml_file.exists(), "experiment.yaml should be created"
    else:
        # If it fails due to missing installation, skip the test
        pytest.skip("AQUA not fully installed or 'ci' catalog not available")


def test_cli_checker_invalid_entry():
    """Test cli_checker with an invalid catalog entry raises NoDataError."""
    checker_script = os.path.join(aqua_pkg_path[0], "core", "analysis", "cli_checker.py")
    
    # Test with an invalid entry
    result = subprocess.run(
        [
            sys.executable, checker_script,
            '--catalog', 'ci',
            '--model', 'invalid-model',
            '--exp', 'invalid-exp',
            '--source', 'invalid-source',
            '--no-read',
            '--no-rebuild',
            '--loglevel', 'WARNING'
        ],
        capture_output=True,
        text=True,
        timeout=30
    )
    
    # Should fail with exit code 1 (or any non-zero)
    assert result.returncode != 0, "cli_checker should fail for invalid entry"
    # Check for error message (may be NoDataError or other Reader initialization error)
    assert "Failed to retrieve data" in result.stderr or "NoDataError" in result.stderr or "not found" in result.stderr.lower()


def test_cli_checker_missing_arguments():
    """Test cli_checker fails when required arguments are missing."""
    checker_script = os.path.join(aqua_pkg_path[0], "core", "analysis", "cli_checker.py")
    
    # Test with missing model
    result = subprocess.run(
        [
            sys.executable, checker_script,
            '--exp', 'test-tco79',
            '--source', 'short',
            '--loglevel', 'WARNING'
        ],
        capture_output=True,
        text=True,
        timeout=10
    )
    
    # Should fail (either ValueError or SystemExit from argparse)
    assert result.returncode != 0
    assert "model, exp and source are required" in result.stderr or "ValueError" in result.stderr




"""AQUA analysis command line interface."""

import os
import sys
import argparse
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from dask.distributed import LocalCluster
from aqua.core.analysis import run_diagnostic_func, run_command, get_aqua_paths
from aqua.core.util import load_yaml, create_folder, format_realization
from aqua.core.configurer import ConfigPath
from aqua.core.util import expand_env_vars
from aqua.core.logger import log_configure
from importlib import resources as pypath


def analysis_parser(parser=None):
    """
    Parser for the AQUA analysis command line interface.
    
    Args:
        parser (argparse.ArgumentParser, optional): An existing parser to extend. If None,
            a new parser will be created.
    
    Returns:
        argparse.ArgumentParser: The configured argument parser.
    """
    if parser is None:
        parser = argparse.ArgumentParser(description="Run AQUA diagnostics.")

    parser.add_argument("-m", "--model", type=str, help="Model (atmospheric and oceanic)")
    parser.add_argument("-e", "--exp", type=str, help="Experiment")
    parser.add_argument("-s", "--source", type=str, help="Source")
    parser.add_argument("--source_oce", type=str, help="Extra source for oceanic data when --source is used for atmospheric data and both are needed")
    parser.add_argument("--realization", type=str, help="Realization (default: None)")
    parser.add_argument("-d", "--outputdir", type=str, help="Output directory")
    parser.add_argument("-f", "--config", type=str, help="Configuration file")
    parser.add_argument("-c", "--catalog", type=str, help="Catalog")
    parser.add_argument("--regrid", type=str, default="False",
                        help="Regrid option (Target grid/False). If False, no regridding will be performed.")
    parser.add_argument("--local_clusters", action="store_true",
                        help="Use separate local clusters instead of single global one")
    parser.add_argument("-p", "--parallel", action="store_true", help="Run diagnostics in parallel with a cluster")
    parser.add_argument("-t", "--threads", type=int, default=-1, help="Maximum number of threads")
    parser.add_argument("--startdate", type=str, help="Start date (YYYY-MM-DD)")
    parser.add_argument("--enddate", type=str, help="End date (YYYY-MM-DD)")
    parser.add_argument("-l", "--loglevel", type=str.upper,
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                        default="INFO", help="Log level")

    return parser

def analysis_execute(args):
    """
    Executing the AQUA analysis by parsing the arguments and configuring the machinery
    """
    loglevel = args.loglevel
    logger = log_configure(loglevel, 'AQUA Analysis')

    aqua_core_path, aqua_diagnostics_path, aqua_configdir, aqua_config_path = get_aqua_paths(args=args, logger=logger)

    config = load_yaml(aqua_config_path)
    loglevel = args.loglevel or config.get('job', {}).get('loglevel', "info")
    logger = log_configure(log_level=loglevel.lower(), log_name='AQUA Analysis')

    model = args.model or config.get('job', {}).get('model')
    exp = args.exp or config.get('job', {}).get('exp')
    source = args.source or config.get('job', {}).get('source', 'lra-r100-monthly')
    source_oce = args.source_oce or config.get('job', {}).get('source_oce')
    realization = args.realization if args.realization else config.get('job', {}).get('realization')

    # startdate and enddate
    startdate = args.startdate or config.get('job', {}).get('startdate')
    enddate = args.enddate or config.get('job', {}).get('enddate')
    # We get regrid option and then we set it to None if it is False
    # This avoids to add the --regrid argument to the command line
    # if it is not needed
    regrid = args.regrid or config.get('job', {}).get('regrid', False)
    if regrid is False or regrid.lower() == 'false':
        regrid = None

    if not all([model, exp, source]):
        logger.error("Model, experiment, and source must be specified either in config or as command-line arguments.")
        sys.exit(1)
    else:
        logger.info(
            "Requested experiment: Model = %s, Experiment = %s, Source = %s. Source_oce = %s",
            model, exp, source, source_oce
        )

    catalog = args.catalog or config.get('job', {}).get('catalog')
    if catalog:
        logger.info("Requested catalog: %s", catalog)
    else:
        cat, _ = ConfigPath().browse_catalogs(model, exp, source)
        if cat:
            catalog = cat[0]
            logger.info("Automatically determined catalog: %s", catalog)
        else:
            logger.error(
                "Model = %s, Experiment = %s, Source = %s triplet not found in any installed catalog.",
                model, exp, source
            )
            sys.exit(1)

    outputdir = os.path.expandvars(args.outputdir or config.get('job', {}).get('outputdir', './output'))
    max_threads = args.threads

    logger.debug("outputdir: %s", outputdir)
    logger.debug("max_threads: %d", max_threads)

    # Format the realization string by prepending 'r' if it is a digit or setting a default `r1`.
    realization = format_realization(realization)
    logger.info("Input realization formatted to: %s", realization)

    output_dir = os.path.join(outputdir, catalog, model, exp, realization)
    output_dir = os.path.expandvars(output_dir)

    # Set Dask timeouts if not already defined in the environment
    if "DASK_DISTRIBUTED__COMM__TIMEOUTS__CONNECT" not in os.environ:
        connect_timeout = config.get('cluster', {}).get('connect_timeout', None)
        if connect_timeout:
            os.environ["DASK_DISTRIBUTED__COMM__TIMEOUTS__CONNECT"] = f"{connect_timeout}s"  # increase timeout (certainly needed on lumi, possibly good anyway)
    if "DASK_DISTRIBUTED__COMM__TIMEOUTS__TCP" not in os.environ:
        tcp_timeout = config.get('cluster', {}).get('tcp_timeout', None)
        if tcp_timeout:
            os.environ["DASK_DISTRIBUTED__COMM__TIMEOUTS__TCP"] = f"{tcp_timeout}s"  # optional, might be good

    os.environ["OUTPUT"] = output_dir
    os.environ["AQUA_CORE"] = aqua_core_path
    os.environ["AQUA_DIAGNOSTICS"] = aqua_diagnostics_path
    os.environ["AQUA_CONFIG"] = aqua_configdir if 'AQUA_CONFIG' not in os.environ else os.environ["AQUA_CONFIG"]
    create_folder(output_dir, loglevel=loglevel)

    # expand the environment variables in the entire config
    config = expand_env_vars(config)

    run_checker = config.get('job', {}).get('run_checker', False)
    if run_checker:
        logger.info("Running setup checker")
        checker_script_path = os.path.join(pypath.files('aqua.core'), 'analysis', 'cli_checker.py')
        output_log_path = os.path.expandvars(f"{output_dir}/setup_checker.log")
        command = f"python {checker_script_path} --model {model} --exp {exp} --source {source} -l {loglevel} --yaml {output_dir}"
        if regrid:
            command += f" --regrid {regrid}"
        if catalog:
            command += f" --catalog {catalog}"
        if realization:
            command += f" --realization {realization}"
        logger.debug("Command: %s", command)
        result = run_command(command, log_file=output_log_path, logger=logger)

        if result == 1:
            logger.critical("Setup checker failed, exiting.")
            sys.exit(1)
        elif result == 0:
            logger.info("Setup checker completed successfully.")
        else:
            logger.error("Setup checker returned exit code %s, check the logs for more information.", result)

    run = config.get('run', [])
    if not run:
        logger.error("No run block found in configuration.")
        sys.exit(1)

    if args.parallel:
        if args.local_clusters:
            logger.info("Running diagnostics in parallel with separate local clusters.")
            cluster = None
            cluster_address = None
        else:
            nthreads = config.get('cluster', {}).get('threads', 2)
            nworkers = config.get('cluster', {}).get('workers', 64)
            mem_limit = config.get('cluster', {}).get('memory_limit', "3.1GiB")

            # silence_logs to avoids excessive logging (see https://github.com/dask/dask/issues/9888)
            cluster = LocalCluster(
                threads_per_worker=nthreads,
                n_workers=nworkers,
                memory_limit=mem_limit,
                silence_logs=logging.ERROR
            )
            cluster_address = cluster.scheduler_address
            logger.info("Initialized global dask cluster %s providing %d workers.", cluster_address, len(cluster.workers))
    else:
        logger.info("Running diagnostics without a dask cluster.")
        cluster = None
        cluster_address = None

    # read cli definitions and prepend script path
    cli = config.get('cli', {})
    script_dir = config.get('job', {}).get("script_path_base")  # we were not using this key
    if script_dir:
        for diag in cli:
            cli[diag] = os.path.join(script_dir, cli[diag])

    # Internal naming scheme:
    # diagnostic: the name of the wrapper metadiagnostic, e.g. atmosphere2d, climate_metrics, etc.
    # tool: the name of the individual command-line tool being run, e.g. biases, ecmean, etc.
    for diag_group in run:

        with ThreadPoolExecutor(max_workers=max_threads if max_threads > 0 else None) as executor:
            futures = []
            for diagnostic in diag_group:

                logger.info("Starting diagnostic: %s", diagnostic)
                diag_config = config.get('diagnostics', {}).get(diagnostic)
                if diag_config is None:
                    logger.error("Diagnostic '%s' not found in the configuration, skipping.", diagnostic)
                    continue

                futures.append(executor.submit(
                    run_diagnostic_func,
                    diagnostic=diagnostic,
                    parallel=args.parallel,
                    diag_config=diag_config,
                    cli=cli,
                    catalog=catalog,
                    model=model,
                    exp=exp,
                    source=source,
                    source_oce=source_oce,
                    realization=realization,
                    startdate=startdate,
                    enddate=enddate,
                    regrid=regrid,
                    output_dir=output_dir,
                    loglevel=loglevel,
                    logger=logger,
                    cluster=cluster_address
                ))

            for future in as_completed(futures):
                try:
                    result = future.result()
                except Exception as e:
                    logger.error("Diagnostic raised an exception: %s", e)

    if cluster:
        cluster.close()
        logger.info("Dask cluster closed.")

    logger.info("All diagnostics finished.")

if __name__ == "__main__":
    args = analysis_parser().parse_args(sys.argv[1:])
    analysis_execute(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA command line main functions
'''

import os
from importlib import resources as pypath

from aqua.core.logger import log_configure

from .parser import parse_arguments
from .analysis import analysis_execute
from .builder import builder_execute
from .drop import drop_execute
from .catgen import catgen_execute
from .install import InstallMixin
from .catalog import CatalogMixin
from .files import FilesMixin

# this are used to check existence of aqua.diagnostics
from .install import DIAGNOSTIC_CONFIG_DIRECTORIES, DIAGNOSTIC_TEMPLATE_DIRECTORIES



class AquaConsole(InstallMixin, CatalogMixin, FilesMixin):
    """Class for AquaConsole, the AQUA command line interface for
    installation, catalog, grids and fixes editing"""

    def __init__(self):
        """The main AQUA command line interface"""

        # NOTE: self.corepath points to $AQUA/aqua/core/config folder
        self.corepath = os.path.join(pypath.files('aqua.core'), 'config')
        if DIAGNOSTIC_CONFIG_DIRECTORIES and DIAGNOSTIC_TEMPLATE_DIRECTORIES:
            self.diagpath = os.path.join(pypath.files('aqua.diagnostics'), 'config')
        else:
            self.diagpath = None
        self.configpath = None
        self.templatepath = None
        self.configfile = 'config-aqua.yaml'
        self.grids = None
        self.logger = None
        self.loglevel = 'WARNING'

        self.command_map = {
            'install': self.install,
            'add': self.add,
            'remove': self.remove,
            'set': self.set,
            'uninstall': self.uninstall,
            'avail': self.avail,
            'list': self.list,
            'update': self.update,
            'fixes': {
                'add': self.fixes_add,
                'remove': self.remove_file
            },
            'grids': {
                'add': self.grids_add,
                'remove': self.remove_file,
                'set': self.grids_set,
                'build': self.grids_build
            },
            'analysis': self.analysis,
            'drop': self.drop,
            'catgen': self.catgen
        }

    def execute(self):
        """Parse AQUA class and run the required command"""

        parser_dict = parse_arguments()
        parser = parser_dict['main']
        args = parser.parse_args()

        # Set the log level
        if args.very_verbose or (args.verbose and args.very_verbose):
            self.loglevel = 'DEBUG'
        elif args.verbose:
            self.loglevel = 'INFO'

        self.logger = log_configure(self.loglevel, 'AQUA')

        command = args.command
        method = self.command_map.get(command, parser.print_help)

        if command not in self.command_map:
            parser.print_help()
        else:
            # nested map
            if isinstance(self.command_map[command], dict):
                if args.nested_command:
                    self.command_map[command][args.nested_command](args)
                else:
                    parser_dict[command].print_help()
            # default
            else:
                method(args)

    def analysis(self, args):
        """
        Run the AQUA analysis

        Args:
            args (argparse.Namespace): arguments from the command line
        """

        print('Running the AQUA analysis')
        analysis_execute(args)

    def drop(self, args):
        """
        Run the Data Reduction OPerator

        Args:
            args (argparse.Namespace): arguments from the command line
        """

        print('Running the Data Reduction OPerator')
        drop_execute(args)

    def catgen(self, args):
        """
        Run the FDB catalog generator

        Args:
            args (argparse.Namespace): arguments from the command line
        """

        print("Running the catalog generator")
        catgen_execute(args)

    def grids_build(self, args):
        """Build grids from data sources

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        print('Running AQUA grids builder')
        builder_execute(args)


def main():
    """AQUA main installation tool"""
    aquacli = AquaConsole()
    aquacli.execute()



"""Reader module."""
from .reader import Reader
from .streaming import Streaming
from .trender import Trender
from .catalog import show_catalog_content

__all__ = ["Reader", "Streaming", "Trender", "show_catalog_content"]



"""Testing if the accessor works"""

import pytest
from aqua import Reader

# Aliases with module scope for fixtures
@pytest.fixture(scope='module')
def reader_instance(ifs_tco79_short_r100_reader):
    return ifs_tco79_short_r100_reader

@pytest.fixture(scope='module')
def data(ifs_tco79_short_r100_data):
    return ifs_tco79_short_r100_data

@pytest.fixture(scope='module')
def reader_instance2(ifs_tco79_short_r200_reader):
    return ifs_tco79_short_r200_reader

@pytest.fixture(scope='module')
def data2(ifs_tco79_short_r200_data):
    return ifs_tco79_short_r200_data

# Test classes
@pytest.mark.aqua
class TestAccessor():
    """Test class for accessor"""

    def test_accessor_fldmean(self, data):
        """Test fldmean as accessor"""
        avg = data.aqua.fldmean()['2t'].values
        assert avg[1] == pytest.approx(285.75920)

    def test_accessor_histogram(self, data):
        """Test histogram as accessor"""
        hist = data['2t'].aqua.histogram(bins=200, range=(150, 350), weighted=False)
        assert sum(hist.values) == data['2t'].size

    def test_accessor_two(self, reader_instance, reader_instance2, data, data2):
        """Test the use of two reader instances"""
        # test accessor on dataset, the object should remember the correct reader
        data1r = data.aqua.regrid()
        data2r = data2.aqua.regrid()

        assert len(data1r.lon) == 360
        assert len(data2r.lon) == 180

        # Check if the accessor still works with the result
        reader_instance.set_default()  # Make sure that the correct reader is used
        avg = data1r.aqua.fldmean()['2t'].values
        assert avg[1] == pytest.approx(285.7543, rel=1e-4)

        # Alternative way to do it
        avg = data2r.aqua.set_default(reader_instance2).aqua.fldmean()['2t'].values
        assert avg[1] == pytest.approx(285.7543, rel=1e-4)

        # Test accessor on dataarray (should be using the previous default)
        avgda = data2['2t'].aqua.fldmean().values
        assert avgda[1] == pytest.approx(285.7543, rel=1e-4)


"""Streaming Mixin for Reader"""

import pandas as pd
import numpy as np
#from aqua.core.logger import log_configure
from aqua.core.util import frequency_string_to_pandas
from aqua.core.util import extract_literal_and_numeric


class Streaming():
    """Streaming class to be used in Reader and elsewhere"""

    def __init__(self, aggregation='S', startdate=None, enddate=None, loglevel=None):
        """
        The Streaming constructor.
        The streamer  is used to stream data by either a specific time interval
        or by a specific number of samples. If the unit parameter is specified, the
        data is streamed by the specified unit and stream_step (e.g. 1 month).
        If the unit parameter is not specified, the data is streamed by stream_step
        steps of the original time resolution of input data.

        If the stream function is called a second time, it will return the subsequent
        chunk of data in the sequence. The function keeps track of the state of the
        streaming process through the use of an internal counter.
        This allows the user to stream through the entire dataset in multiple calls to the function,
        retrieving consecutive chunks of data each time.

        If startdate is not specified, the method will use the first date in the dataset.

        Arguments:
            startdate (str): the starting date for streaming the data (e.g. '2020-02-25') (None)
            enddate (str): the ending date for streaming the data (e.g. '2021-01-01') (None)
            aggregation (str): the streaming frequency in pandas style (1M, 7D etc.)
            loglevel (string):      Level of logging according to logging module
                                    (default: log_level_default of loglevel())

        Returns:
            A `Streaming` class object.
        """

        # define the internal logger
        # self.logger = log_configure(log_level=loglevel, log_name='Streaming')
        self.startdate = startdate
        self.enddate = enddate
        self.aggregation = aggregation
        self.idx = 0

    def stream_chunk(self, data, startdate=None, enddate=None, aggregation=None):
        """
        Compute chunks for a dataset using startdate, enddate and aggregation defined by the constructor.

        Arguments:
            data (xr.Dataset):      the input xarray.Dataset
            startdate (str): the starting date for streaming the data (e.g. '2020-02-25') (None)
            enddate (str): the ending date for streaming the data (e.g. '2021-01-01') (None)
            aggregation (str): the streaming frequency in pandas style (1M, 7D etc.)

        Returns:
            A DataArrayResample object for the time axis
        """

        if not startdate:
            startdate = self.startdate
        if not enddate:
            enddate = self.enddate
        if not enddate:  # If it is still not defined use the end of the data
            enddate = data.time[-1].values
        if not aggregation:
            aggregation = self.aggregation
        if not aggregation:  # if it is still None set to step
            aggregation = "S"

        aggregation = frequency_string_to_pandas(aggregation)

        if startdate:
            tim = data.time.sel(time=slice(startdate, enddate))
        else:
            tim = data.time

        literal, numeric = extract_literal_and_numeric(aggregation)        

        if literal == 'S':
            #nsteps = np.maximum(int('0' + numeric), 1)  # this allows also "S" for "1S"
            timr = pd.Series(tim).groupby(by=(np.arange(0, len(tim)) // numeric))
        else:
            timr = tim.resample(time=aggregation)

        return timr

    def stream(self, data, startdate=None, enddate=None, aggregation=None,
               timechunks=None, reset=False):
        """
        Stream a chunk of a dataset using startdate, enddate and aggregation defined by the constructor.

        Arguments:
            data (xr.Dataset):      the input xarray.Dataset
            startdate (str): the starting date for streaming the data (e.g. '2020-02-25') (None)
            enddate (str): the ending date for streaming the data (e.g. '2021-01-01') (None)
            aggregation (str): the streaming frequency in pandas style (1M, 7D etc.)
            timechunks (DataArrayResample, optional): a precomputed chunked time axis
            reset (bool, optional): reset the streaming

        Returns:
            A xarray.Dataset containing the subset of the input data that has been streamed.
        """

        if reset:
            self.idx = 0

        if not timechunks:
            timechunks = self.stream_chunk(data, startdate=startdate, enddate=enddate, aggregation=aggregation)

        if self.idx >= len(timechunks):  # we have consumed all the data
            return None
        else:
            date1 = timechunks.first()[self.idx]
            date2 = timechunks.last()[self.idx]
            self.idx += 1
            return (data.sel(time=slice(date1, date2)))

    def reset(self):
        """
        Reset the state of the streaming process.
        This means that if the stream function is called again after calling reset_stream,
        it will start streaming the input data from the beginning.
        """
        self.idx = 0



import os
import pytest
from pathlib import Path
import tempfile
import subprocess
from aqua.core.util.checksum import compute_md5, generate_checksums, verify_checksums


@pytest.mark.aqua
class TestChecksumModule:
    def test_compute_md5(self, tmp_path):
        """Test MD5 checksum computation using tmp_path."""
        file = tmp_path / "test.txt"
        file.write_text("Test content")

        md5 = compute_md5(str(file))
        
        assert md5 is not None
        assert len(md5) == 32  # MD5 checksum should be 32 characters long

    def test_compute_md5_file_not_found(self):
        """Test compute_md5 for a non-existent file."""
        md5 = compute_md5("non_existent_file.txt")
        assert md5 is None

    def test_full_checksums(self, tmp_path, capsys):
        """Test checksum generation."""
        file1 = tmp_path / "file1.txt"
        file2 = tmp_path / "file2.txt"
        file1.write_text("Sample content 1")
        file2.write_text("Sample content 2")

        checksum_file = Path("checksums.md5")
        generate_checksums(tmp_path, ["."],str(checksum_file))

        assert checksum_file.exists()
        with checksum_file.open("r", encoding='utf8') as f:
            content = f.read()

        verify_checksums(tmp_path, ["."], str(checksum_file))
        captured = capsys.readouterr()
        assert "All files are verified successfully" in captured.out

        os.remove(checksum_file)






"""New Regrid class independent from the Reader"""
import os
import re
import shutil
from tempfile import TemporaryDirectory
import xarray as xr
from smmregrid import CdoGenerate, GridInspector
from smmregrid import Regridder as SMMRegridder
from smmregrid.util import check_gridfile
from aqua.core.logger import log_configure
from aqua.core.util import to_list
from .griddicthandler import GridDictHandler
from .regridder_util import check_existing_file, validate_reader_kwargs


# parameters which will affect the weights and areas name
DEFAULT_WEIGHTS_AREAS_PARAMETERS = ['zoom']

# default CDO regrid method
DEFAULT_GRID_METHOD = 'ycon'

# default dimension for the weights and areas
DEFAULT_DIMENSION = '2d'
DEFAULT_DIMENSION_MASK = '2dm'  # masked grid

# please notice: check_gridfile is a function from smmregrid.util
# to check and if a grid is a cdo grid,
# a file on the disk or xarray dataset. Possible inclusion of CDOgrid object is considered
# but should be likely developed on the smmregrid side.


class Regridder():
    """AQUA Regridder class"""

    def __init__(self, cfg_grid_dict: dict = None,
                 src_grid_name: str = None,
                 data: xr.Dataset = None,
                 cdo: str = None,
                 loglevel: str = "WARNING"):
        """
        The (new) Regridder class. Can be initialized with a data (xr.Dataset/DataArray) or a src_grid_name
        It provides methods to generate areas and weights, and to regrid a dataset.

        Args:
            cfg_grid_dict (dict): The dictionary containing the full AQUA grid configuration.
            src_grid_name (str, optional): The name of the source grid in the AQUA convention.
            data (xarray.Dataset, optional): The dataset to be regridded if src_grid_name is not provided.
            cdo (str, optional): The path to the CDO executable. If None, guess it from the system.
            loglevel (str): The logging level.

        Attributes:
            loglevel (str): The logging level.
            logger (logging.Logger): The logger.
            cfg_grid_dict (dict): The full AQUA grid dictionary.
            src_grid_name (str): The source grid name.
            handler (GridDictHandler): The grid dictionary handler.
            src_grid_dict (dict): The normalized source grid dictionary.
            src_horizontal_dims (str): The source horizontal dimensions.
            src_vertical_dim (str): The source vertical dimension.
            tgt_horizontal_dims (str): The target horizontal dimensions.
            error (str): The error message to be used by the Reader.
            cdo (str): The CDO path.
            smmregridder (dict): The SMMregrid regridder object for each vertical coordinate.
            src_grid_area (xarray.Dataset): The source grid area.
            tgt_grid_area (xarray.Dataset): The target grid area.
            masked_attrs (dict): The masked attributes.
            masked_vars (list): The masked variables.
            extra_dims (dict): The extra dimensions (from cfg_grid_dict) to be sent to smmregrid.
        """

        if src_grid_name is None and data is None:
            raise ValueError("Either src_grid_name or data must be provided.")

        self.loglevel = loglevel
        self.logger = log_configure(log_level=loglevel, log_name='Regridder')

        # define basic attributes:
        self.cfg_grid_dict = cfg_grid_dict if cfg_grid_dict else {}  # full grid dictionary
        self.src_grid_name = src_grid_name  # source grid name

        # we want all the grid dictionary to be real dictionaries
        self.handler = GridDictHandler(cfg_grid_dict,
                                       default_dimension=DEFAULT_DIMENSION,
                                       loglevel=loglevel)
        self.src_grid_dict = self.handler.normalize_grid_dict(self.src_grid_name)
        self.src_grid_path = self.src_grid_dict.get('path')

        self.logger.debug("Normalized grid dictionary: %s", self.src_grid_dict)
        self.logger.debug("Normalized grid path: %s", self.src_grid_path)

        # this not used but can be shipped back to the reader
        self.src_horizontal_dims = self.src_grid_dict.get('space_coord', None)
        self.src_vertical_dim = list(self.src_grid_path.keys())
        self.tgt_horizontal_dims = None
        self.error = None

        self.logger.debug("Horizontal dimensions: %s", self.src_horizontal_dims)
        self.logger.debug("Vertical dimensions: %s", self.src_vertical_dim)

        # store dimension to be send to smmregrid if needed
        self.extra_dims = {
            'vertical': to_list(self.src_vertical_dim),
            'horizontal': to_list(self.src_horizontal_dims)
        }

        # check data to extract information for CDO
        if data is not None:
            self._get_info_from_data(data)

        if not self.src_horizontal_dims:
            self.error = "Horizontal dimensions not found in the grid path. Please provide a dataset. "
            self.logger.warning(self.error)
            return

        # check if the grid path has been defined
        if not self.src_grid_path:
            self.error = "Source grid path not found. Please provide a dataset."
            self.logger.warning(self.error)
            return

        # check if CDO is available
        self.cdo = self._set_cdo(cdo=cdo)

        # SMMregridders dictionary for each vertical coordinate
        self.smmregridder = {}

        # source and target areas
        self.src_grid_area = None
        self.tgt_grid_area = None

        # configure the masked fields
        self.masked_attrs, self.masked_vars = self.configure_masked_fields(self.src_grid_dict)

        self.logger.info("Grid name: %s", self.src_grid_name)
        self.logger.debug("Grid dictionary: %s", self.src_grid_dict)

    def _set_cdo(self, cdo=None):
        """
        Check information on CDO to set the correct version.

        Args:
            cdo (str, optional): The path to the CDO executable. If None, guess it from the system.

        Returns:
            str: The path to the CDO executable.

        Raises:
            FileNotFoundError: If CDO is not found in the system path.
        """
        if cdo:
            # TODO: add a subprocess call to add if cdo is available
            self.logger.debug("Going to use CDO in: %s", cdo)
            return cdo

        cdo = shutil.which("cdo")
        if cdo:
            self.logger.debug("Found CDO path: %s", cdo)
            return cdo

        raise FileNotFoundError(
                "CDO not found in path: Weight and area generation will fail.")

    def _get_info_from_data(self, data):
        """
        Extract information from the dataset to be used in the regridding process
        """

        gridinspector = GridInspector(data, loglevel=self.loglevel, extra_dims=self.extra_dims)
        gridtypes = gridinspector.get_gridtype()

        # if we have not them from the dictionary, get it from the file
        if not self.src_horizontal_dims:
            # however we have only one 2d grid at the time in AQUA
            self.src_horizontal_dims = gridtypes[0].horizontal_dims
        self.logger.debug("Horizontal dimensions guessed from data: %s", self.src_horizontal_dims)

        # This should be not necessary since vertical coordinate is always provided
        # if we have not them from the dictionary, get it from the file
        # if not self.src_vertical_dim:
        #    # get all vertical grid available
        #    self.src_vertical_dim = [getattr(gridtype, "vertical_dim") for gridtype in gridtypes]
        # self.logger.debug("Vertical dimensions guessed from data: %s", self.src_vertical_dim)

        # if the path is missing, use the data to init smmregrid
        if not self.src_grid_path:
            vdim = self.src_vertical_dim if self.src_vertical_dim else DEFAULT_DIMENSION
            self.logger.info("Using provided dataset as a grid path for %s", vdim)
            self.src_grid_dict = {"path": {vdim: data}}
            self.src_grid_path = self.src_grid_dict.get('path')

    def areas(self, tgt_grid_name=None, rebuild=False, reader_kwargs=None):
        """
        Load or generate regridding areas for the source or target grid.

        Args:
            tgt_grid_name (str, optional): Name of the target grid.
                                           If None, the self.src_grid_name is used.
            rebuild (bool, optional): If True, forces regeneration of the area.
            reader_kwargs (dict, optional): Additional parameters for the reader.

        Returns:
            xr.Dataset: The computed grid area.
        """

        # normalize dictionaries for target grid
        if tgt_grid_name:
            grid_dict = self.handler.normalize_grid_dict(tgt_grid_name)
        else:
            grid_dict = self.src_grid_dict

        # generate the area
        grid_area = self._load_area(
            grid_name=tgt_grid_name,
            grid_dict=grid_dict,
            reader_kwargs=reader_kwargs,
            rebuild=rebuild
        )

        # assign the area to the correct attribute
        if tgt_grid_name:
            self.tgt_grid_area = grid_area
            # Extra: infer target grid horizontal dimensions
            self.tgt_horizontal_dims = GridInspector(
                self.tgt_grid_area, loglevel=self.loglevel
            ).get_gridtype()[0].horizontal_dims

        else:
            self.src_grid_area = grid_area

        return grid_area
    
    def _safe_to_netcdf(self, data, filename):
        """Save to netcdf safely using a temporary file.
        
        Args:
            data (xr.Dataset or xr.DataArray): Data to save.
            filename (str): Destination file path.
        """
        # Ensure parent directory exists
        dest_dir = os.path.dirname(os.path.abspath(filename))
        if dest_dir:  # Handle edge case where filename has no directory component
            os.makedirs(dest_dir, exist_ok=True)
        else:
            dest_dir = '.'  # Use current directory
        
        # Create temp file in same directory as destination (same filesystem)
        with TemporaryDirectory(dir=dest_dir) as tmpdirname:
            tmp_file = os.path.join(tmpdirname, "temp.nc")
            data.to_netcdf(tmp_file)
            os.replace(tmp_file, filename)

    def _load_area(self, grid_name, grid_dict, reader_kwargs, rebuild=False):
        """
        Load or generate the grid area.

        Args:
            grid_name (str): The grid name. If None, roll back to src_grid_name.
            grid_dict (dict): The normalized grid dictionary.
            reader_kwargs (dict): Additional reader parameters.
            target (bool): Whether this is for the target grid (default: False).
            rebuild (bool): If True, forces regeneration of the area.

        Returns:
            xr.Dataset: The computed grid area.
        """
        area_filename = self._area_filename(grid_name if grid_name else None, reader_kwargs)
        area_type = "target" if grid_name else "source"

        # if file exists, load it
        if not rebuild and check_existing_file(area_filename):
            self.logger.info("Loading existing %s area from %s.", area_type, area_filename)
            return xr.open_dataset(area_filename)

        # generate and save the area
        grid_area = self._generate_area(grid_name, grid_dict, area_filename, area_type)
        self._safe_to_netcdf(grid_area, area_filename)
        self.logger.info("Saved %s area to %s.", area_type, area_filename)

        return grid_area

    def _generate_area(self, grid_name, grid_dict, area_filename, area_type):
        """
        Loads cell areas if available; otherwise, generates the area.

        Args:
            grid_name (str): The grid name: if None, the source grid is used.
            grid_dict (dict): The normalized grid dictionary.
            area_filename (str): The precomputed area filename.
            area_type (str): The area type (i.e. source or target)
        """

        # if they have been provided, read from the AQUA dict
        cellareas, cellareas_var = grid_dict.get('cellareas'), grid_dict.get('cellareas_var')
        if cellareas and cellareas_var:
            self.logger.info("Using cellareas from variable %s in file %s",
                             cellareas_var, cellareas)
            if not os.path.exists(cellareas):
                raise FileNotFoundError(f"Grid based cell area  file {cellareas} not found.")
            return xr.open_mfdataset(cellareas)[cellareas_var].rename("cell_area").squeeze().to_dataset()

        # clean if necessary
        if os.path.exists(area_filename):
            self.logger.info("%s areas file %s exists. Regenerating.", area_type, area_filename)

        self.logger.info("Generating %s area for %s", area_type, grid_name)

        source_grid = self._get_grid_path(grid_dict.get('path')) if area_type == "source" else None
        target_grid = self._get_grid_path(grid_dict.get('path')) if area_type == "target" else None

        return CdoGenerate(
            source_grid=source_grid,
            target_grid=target_grid,
            cdo_extra=grid_dict.get('cdo_extra'),
            cdo_options=grid_dict.get('cdo_options'),
            cdo=self.cdo,
            loglevel=self.loglevel
        ).areas(target=bool(grid_name))

    def weights(self, tgt_grid_name, regrid_method=None, nproc=1,
                rebuild=False, reader_kwargs=None):
        """
        Load or generate regridding weights calling smmregrid

        Args:
            tgt_grid_name (str): The destination grid name.
            regrid_method (str): The regrid method.
            nproc (int): The number of processors to use.
            rebuild (bool): If True, rebuild the weights.
            reader_kwargs (dict): The reader kwargs for filename definition,
                                  including info on model, exp, source, etc.
        """

        # define regrid method
        default_regrid_method = self.src_grid_dict.get('regrid_method', DEFAULT_GRID_METHOD)
        regrid_method = regrid_method if regrid_method else default_regrid_method
        if regrid_method is not DEFAULT_GRID_METHOD:
            self.logger.info("Regrid method: %s", regrid_method)

        # normalize the tgt grid dictionary and path
        tgt_grid_dict = self.handler.normalize_grid_dict(tgt_grid_name)

        # get the cdo options from the configuration
        cdo_extra = self.src_grid_dict.get('cdo_extra', None)
        cdo_options = self.src_grid_dict.get('cdo_options', None)

        # loop over the vertical coordinates: DEFAULT_DIMENSION, DEFAULT_DIMENSION_MASK, or any other
        for vertical_dim in self.src_grid_path:

            # define the vertical coordinate in the smmregrid world
            smm_vertical_dim = None if vertical_dim in [
                DEFAULT_DIMENSION, DEFAULT_DIMENSION_MASK] else vertical_dim

            weights_filename = self._weights_filename(tgt_grid_name, regrid_method,
                                                      vertical_dim, reader_kwargs)

            # check if weights already exist, if not, generate them
            if rebuild or not check_existing_file(weights_filename):

                if os.path.exists(weights_filename):
                    self.logger.info(
                        "Weights file %s exists. Regenerating.", weights_filename)
                else:
                    self.logger.info(
                        "Generating weights for %s grid: %s", tgt_grid_name, vertical_dim)

                if smm_vertical_dim:
                    self.logger.warning("Mask-changing vertical dimension identified, weights generation might take a few!")

                # smmregrid call
                # TODO: here or better in smmregird, we could use GridInspect to get the grid info
                # and reduce the dimensionality of the input data.
                generator = CdoGenerate(source_grid=self.src_grid_path[vertical_dim],
                                        target_grid=self._get_grid_path(tgt_grid_dict.get('path')),
                                        cdo_extra=cdo_extra,
                                        cdo_options=cdo_options,
                                        cdo=self.cdo,
                                        loglevel=self.loglevel)

                # generate and save the weights
                weights = generator.weights(method=regrid_method,
                                            vertical_dim=smm_vertical_dim,
                                            nproc=nproc)
                self._safe_to_netcdf(weights, weights_filename)

            else:
                self.logger.info(
                    "Loading existing weights from %s.", weights_filename)
                
            # load the weights
            weights = xr.open_dataset(weights_filename)

            # initialize the regridder
            self.smmregridder[vertical_dim] = SMMRegridder(
                weights=weights,
                horizontal_dims=self.src_horizontal_dims,
                vertical_dim=smm_vertical_dim,
                loglevel=self.loglevel
            )

    def _area_filename(self, tgt_grid_name, reader_kwargs):
        """"
        Generate the area filename.

        Args:
            tgt_grid_name (str): The destination grid name.
            reader_kwargs (dict): The reader kwargs, including info on model, exp, source, etc.
        """

        area_dict = self.cfg_grid_dict.get('areas')

        if not area_dict:
            self.logger.warning(
                "Areas block not found in the configuration file, using fallback naming scheme.")
            if tgt_grid_name:
                return f"cell_area_{tgt_grid_name}.nc"
            return f"cell_area_{self.src_grid_name}.nc"

        # destination grid name is provided, use grid template
        if tgt_grid_name:
            filename = area_dict["template_grid"].format(grid=tgt_grid_name)
            self.logger.debug(
                "Using grid-based template for target grid. Filename: %s", filename)
        # source grid name is provided, check if it is data
        else:
            if check_gridfile(self._get_grid_path(self.src_grid_path)) != 'xarray':
                filename = area_dict["template_grid"].format(
                    grid=self.src_grid_name)
                self.logger.debug(
                    "Using grid-based template for source grid. Filename: %s", filename)
            else:
                reader_kwargs = validate_reader_kwargs(reader_kwargs)
                filename = area_dict["template_default"].format(
                    model=reader_kwargs["model"],
                    exp=reader_kwargs["exp"],
                    source=reader_kwargs["source"])
                self.logger.debug(
                    "Using source-based template for source grid. Filename: %s", filename)

        filename = self._insert_kwargs(filename, reader_kwargs)
        filename = self._filename_prepend_path(filename, kind="areas")
        return filename

    def _weights_filename(self, tgt_grid_name, regrid_method, vertical_dim, reader_kwargs):
        """
        Generate the weights filename.

        Args:
            tgt_grid_name (str): The destination grid name.
            regrid_method (str): The regrid method.
            vertical_dim (str): The vertical dimension.
            reader_kwargs (dict): The reader kwargs, including info on model, exp, source, etc.
        """

        levname = vertical_dim if vertical_dim in [
            DEFAULT_DIMENSION, DEFAULT_DIMENSION_MASK] else f"3d-{vertical_dim}"

        weights_dict = self.cfg_grid_dict.get('weights')

        if not weights_dict:
            self.logger.warning(
                "Weights block not found in the configuration file, using fallback naming scheme.")
            return f"weights_{tgt_grid_name}_{regrid_method}_l{levname}.nc"

        # destination grid name is provided, use grid template
        if check_gridfile(self.src_grid_path[vertical_dim]) != 'xarray':
            filename = weights_dict["template_grid"].format(
                sourcegrid=self.src_grid_name,
                method=regrid_method,
                targetgrid=tgt_grid_name,
                level=levname)
            self.logger.debug(
                "Using grid-based template for target grid. Filename: %s", filename)
        else:
            reader_kwargs = validate_reader_kwargs(reader_kwargs)
            filename = weights_dict["template_default"].format(
                model=reader_kwargs["model"],
                exp=reader_kwargs["exp"],
                source=reader_kwargs["source"],
                method=regrid_method,
                targetgrid=tgt_grid_name,
                level=levname)
            self.logger.debug(
                "Using source-based template for target grid. Filename: %s", filename)

        filename = self._insert_kwargs(filename, reader_kwargs)
        filename = self._filename_prepend_path(filename, kind="weights")
        return filename

    def _filename_prepend_path(self, filename, kind="weights"):
        """
        Prepend the path to the filename with some fall back option
        """
        if not self.cfg_grid_dict.get("paths"):
            self.logger.warning(
                "Paths block not found in the configuration file, using present directory.")
        else:
            if not self.cfg_grid_dict["paths"].get(kind):
                self.logger.warning(
                    "%s block not found in the paths block, using present directory.", kind)
            else:
                # if path does not exist, create it
                if not os.path.exists(self.cfg_grid_dict["paths"][kind]):
                    self.logger.warning(
                        "%s path in %s does not exist: creating!", kind, self.cfg_grid_dict["paths"][kind])
                    os.makedirs(self.cfg_grid_dict["paths"][kind], exist_ok=True)
                filename = os.path.join(
                    self.cfg_grid_dict["paths"][kind], filename)
        return filename

    def _expand_dims(self, data, vertical_dims):
        """
        Expand the dimensions of the dataset or dataarray to include the vertical dimensions
        """

        if not list(set(data.dims) & set(vertical_dims)):
            for vertical_dim in vertical_dims:
                if vertical_dim in data.coords:
                    self.logger.debug(
                        "Expanding dimensions to include %s", vertical_dim)
                    data = data.expand_dims(dim=vertical_dim, axis=0)
        return data

    def _group_shared_dims(self, data):
        """
        Groups variables in a dataset that share the same AQUA vertical dimension.
        Built on GridInspector and GridType classes from smmregrid.
        It is a sort of overkill of what smmregrid do internally.

        Args:
            data (xarray.Dataset): The dataset to be regridded.

        Return:
            shared_vars (dict): A dictionary of variables that share the same vertical dimensions.
        """

        shared_vars = {}
        # TODO: masked vars based on attributes are still missing
        if self.masked_vars:
            shared_vars[DEFAULT_DIMENSION_MASK] = self.masked_vars
            self.logger.debug("Variables for coordinate %s: %s",
                              DEFAULT_DIMENSION_MASK, shared_vars[DEFAULT_DIMENSION_MASK])
        # Get the masked variables safely
        masked_vars = shared_vars.get(DEFAULT_DIMENSION_MASK, [])

        # scan the grid
        gridtypes = GridInspector(data, extra_dims=self.extra_dims, loglevel=self.loglevel).get_gridtype()

        self.logger.debug("Gridtypes found: %s", len(gridtypes))
        for gridtype in gridtypes:
            variables = list(gridtype.variables.keys())

            if gridtype.vertical_dim:
                self.logger.debug("Variables for dimension %s: %s",
                                  gridtype.vertical_dim, variables)
                shared_vars[gridtype.vertical_dim] = [var for var in variables if var not in masked_vars]
            else:
                shared_vars[DEFAULT_DIMENSION] = [var for var in variables if var not in masked_vars]
                self.logger.debug("Variables for dimensions %s: %s",
                                  DEFAULT_DIMENSION, shared_vars[DEFAULT_DIMENSION])

        return shared_vars

    def regrid(self, data):
        """
        Actual regridding core function. Regrid the dataset or dataarray using common gridtypes
        Firstly, expand the dimensions of the dataset to include the vertical dimensions if necessary.
        Then, group variables that share the same dimensions.
        Finally, apply regridding on the different vertical coordinates, including 2d and 2dm.

        Args:
            data (xarray.Dataset, xarray.DataArray): The dataset to be regridded.
        """

        # expand the dimensions of the dataset to include the vertical dimensions
        if isinstance(data, xr.Dataset):
            data = data.map(self._expand_dims, vertical_dims=list(self.src_vertical_dim))
        elif isinstance(data, xr.DataArray):
            data = self._expand_dims(data, vertical_dims=list(self.src_vertical_dim))
        else:
            raise ValueError("Data must be an xarray Dataset or DataArray.")

        # get which variables share the same dimensions
        shared_vars = self._group_shared_dims(data)

        # compact regridding on all dataset with map
        if isinstance(data, xr.Dataset):
            data = data.map(self._apply_regrid, shared_vars=shared_vars)
        elif isinstance(data, xr.DataArray):
            data = self._apply_regrid(data, shared_vars)

        return data

    def _apply_regrid(self, data, shared_vars):
        """
        Core regridding function.
        Apply regridding on the different vertical coordinates, including 2d and 2dm
        """

        for vertical, variables in shared_vars.items():
            if data.name in variables:
                if not self.smmregridder.get(vertical):
                    self.logger.error("Regridder for vertical coordinate %s not found.", vertical)
                    self.logger.error("Cannot regrid variable %s", data.name)
                    continue
                # TODO: if smmregridder is not found, we can call the weights method to generate on the fly
                return self.smmregridder[vertical].regrid(data)
        return data

    @staticmethod
    def _insert_kwargs(filename, reader_kwargs):
        """
        Insert the DEFAULT_WEIGHTS_AREAS_PARAMETERS in the filename template.
        """
        # add the kwargs naming in the template file
        if isinstance(reader_kwargs, dict):
            for parameter in DEFAULT_WEIGHTS_AREAS_PARAMETERS:
                if parameter in reader_kwargs:
                    filename = re.sub(
                        r'\.nc', '_' + parameter +
                        str(reader_kwargs[parameter]) + r'\g<0>',
                        filename)

        return filename

    @staticmethod
    def configure_masked_fields(src_grid_dict):
        """
        if the grids has the 'masked' option, this can be based on
        generic attribute or alternatively of a series of specific variables using the 'vars' key

        Args:
            source_grid (dict): Dictionary containing the grid information

        Returns:
            masked_attr (dict): Dict with name and proprierty of the attribute to be used for masking
            masked_vars (list): List of variables to mask
        """
        masked_info = src_grid_dict.get("masked")
        if masked_info is None:
            return None, None

        masked_vars = masked_info.get("vars")
        masked_attrs = {k: v for k, v in masked_info.items() if k !=
                        "vars"} or None

        return masked_attrs, masked_vars

    @staticmethod
    def _get_grid_path(grid_path):
        """
        Get the grid path from the grid dictionary.
        This looks for `DEFAULT_DIMENSION` key,
        otherwise takes the first available value in the dict.
        """
        return grid_path.get(DEFAULT_DIMENSION, next(iter(grid_path.values()), None))



import os
import xarray as xr
from metpy.units import units
from aqua.core.logger import log_configure, log_history
from aqua.core.configurer import ConfigPath
from .yaml import load_yaml


def normalize_units(src, loglevel='WARNING'):
    """
    Get rid of stange grib units based on the default.yaml fix file

    Arguments:
        src (str): input unit to be fixed
    """
    logger = log_configure(loglevel, 'normalize_units')
    src = str(src)

    config_folder = ConfigPath().get_config_dir()
    config_folder = os.path.join(config_folder, "fixes")
    default_file = os.path.join(config_folder, "default.yaml")

    if not os.path.exists(default_file):
        raise FileNotFoundError(f"Cannot find default.yaml in {config_folder}")

    default_dict = load_yaml(default_file)
    fix_units = default_dict['defaults']['units']['fix']
    for key in fix_units:
        if key == src:
            # return fixed
            logger.info('Replacing non-metpy unit %s with %s', key, fix_units[key])
            return src.replace(key, fix_units[key])

    # return original
    return src


def convert_units(src, dst, deltat=None, var="input var", loglevel='WARNING'):
    """
    Converts source to destination units using metpy.
    Returns a dictionary with conversion factors and offsets.

    Arguments:
        src (str): Source units.
        dst (str): Destination units.
        deltat (float, optional): Time delta in seconds (needed for some unit conversions).
        var (str): Variable name (optional, used only for diagnostic output).
        loglevel (str): Log level for the logger. Default is 'WARNING'.

    Returns:
        dict: A dictionary with keys `factor`, `offset`, and possible extra flags
              (e.g., `time_conversion_flag`).
    """
    logger = log_configure(loglevel, 'convert_units')
    src = normalize_units(src, loglevel)
    dst = normalize_units(dst, loglevel)
    factor = units(src).to_base_units() / units(dst).to_base_units()

    # Dictionary for storing conversion attributes
    conversion = {}

    # Flag for time-dependent conversions
    if "second" in str(factor.units) and deltat is not None:
        conversion['time_conversion_flag'] = 1
        conversion['deltat'] = str(deltat)
    elif "second" in str(factor.units) and deltat is None:
        logger.warning("Time-dependent conversion factor detected, but no accumulation time provided")

    if factor.units == units('dimensionless'):
        offset = (0 * units(src)).to(units(dst)) - (0 * units(dst))
    else:
        if factor.units == "meter ** 3 / kilogram":
            factor *= 1000 * units("kg m-3")
            if logger:
                logger.debug("%s: corrected multiplying by density of water 1000 kg m-3", var)
        elif factor.units == "meter ** 3 * second / kilogram":
            factor *= 1000 * units("kg m-3") / (deltat * units("s"))
            if logger:
                logger.debug("%s: corrected multiplying by density of water 1000 kg m-3", var)
                logger.info("%s: corrected dividing by accumulation time %s s", var, deltat)
        elif factor.units == "second":
            factor /= deltat * units("s")
            if logger:
                logger.debug("%s: corrected dividing by accumulation time %s s", var, deltat)
        elif factor.units == "kilogram / meter ** 3":
            factor /= 1000 * units("kg m-3")
            if logger:
                logger.debug("%s: corrected dividing by density of water 1000 kg m-3", var)
        else:
            if logger:
                logger.debug("%s: incommensurate units converting %s to %s --> %s",
                             var, src, dst, factor.units)
        offset = 0 * units(dst)

    # Store non-default conversion factors and offsets
    if offset.magnitude != 0:
        conversion['offset'] = offset.magnitude
    elif factor.magnitude != 1:
        conversion['factor'] = factor.magnitude

    return conversion


def convert_data_units(data, var: str, units: str, loglevel: str = 'WARNING'):
    """
    Converts in-place the units of a variable in an xarray Dataset or DataArray.

    Args:
        data (xarray Dataset or DataArray): The data to be checked.
        var (str): The variable to be checked.
        units (str): The units to be checked.
    """
    logger = log_configure(log_name='check_data', log_level=loglevel)

    data_to_fix = data[var] if isinstance(data, xr.Dataset) else data
    final_units = units
    initial_units = data_to_fix.units

    conversion = convert_units(initial_units, final_units)

    factor = conversion.get('factor', 1)
    offset = conversion.get('offset', 0)

    if factor != 1 or offset != 0:
        logger.debug('Converting %s from %s to %s',
                     var, initial_units, final_units)
        data_to_fix = data_to_fix * factor + offset
        data_to_fix.attrs['units'] = final_units
        log_history(data_to_fix, f"Converting units of {var}: from {initial_units} to {final_units}")
    else:
        logger.debug('Units of %s are already in %s', var, final_units)
        return data

    if isinstance(data, xr.Dataset):
        data_fixed = data.copy()
        data_fixed[var] = data_to_fix
    else:
        data_fixed = data_to_fix

    return data_fixed


def multiply_units(unit1: str, unit2: str, normalise_units=True, 
                   to_base_units=True, loglevel: str = 'WARNING') -> str:
    """
    Multiply two unit strings together using metpy.
    
    This is useful when combining physical quantities. For example, integrating 
    a field in meters (e.g. sea ice thickness) over an area in m**2 (e.g. areacello), 
    resulting in m**3 (e.g. sea ice volume)).
    
    Args:
        unit1 (str): First unit string (e.g., 'm')
        unit2 (str): Second unit string (e.g., 'm2')
        normalise_units (bool): If True, normalize the units. Default is True.
        to_base_units (bool): If True, convert the units to base units. Default is True.
        loglevel (str): Log level for the logger. Default is 'WARNING'.
    
    Returns:
        str: The multiplied unit string (e.g., 'm**3')
        
    Example:
        >>> multiply_units('m', 'm2')
        'm**3'
    """
    unit1 = normalize_units(unit1, loglevel) if normalise_units else unit1
    unit2 = normalize_units(unit2, loglevel) if normalise_units else unit2
    
    result = units(unit1) * units(unit2)
    
    result = result.to_base_units() if to_base_units else result
    return str(result.units)


import os
import time
import threading
from filelock import SoftFileLock, Timeout

from aqua.core.logger import log_configure

class SafeFileLock:
    """
    A resilient soft lock with:
      - Timeout on acquire
      - Heartbeat updates (refreshes lock timestamp periodically)
      - Stale lock cleanup
    """

    def __init__(self, lock_path, timeout=60, stale_timeout=120,
                 heartbeat_interval=10, loglevel: str = 'WARNING'):
        """
        :param lock_path: Path to the .lock file (must be on shared filesystem)
        :param timeout: Seconds to wait to acquire lock before raising Timeout
        :param stale_timeout: Lock older than this (in seconds) is considered stale
        :param heartbeat_interval: How often to refresh lock file's mtime (seconds)
        :param loglevel: Logging level for the lock (DEBUG, INFO, WARNING, ERROR)
        """
        self.lock_path = lock_path
        self.timeout = timeout
        self.stale_timeout = stale_timeout
        self.heartbeat_interval = heartbeat_interval

        self.logger = log_configure(log_level=loglevel, log_name='FileLock')

        self.lock = SoftFileLock(lock_path, timeout=timeout)
        self._stop_event = threading.Event()
        self._heartbeat_thread = None

    def _is_stale(self):
        """Return True if lock file is older than stale_timeout."""
        if not os.path.exists(self.lock_path):
            return False
        age = time.time() - os.path.getmtime(self.lock_path)
        return age > self.stale_timeout

    def _remove_stale(self):
        """Remove stale lock file if too old."""
        if self._is_stale():
            try:
                os.remove(self.lock_path)
                self.logger.debug("Removed stale lock: %s", self.lock_path)
            except OSError:
                pass

    def _write_metadata(self):
        """Write PID + timestamp inside the lock file."""
        try:
            with open(self.lock_path, "w") as f:
                f.write(f"pid={os.getpid()} time={time.time()}\n")
        except Exception:
            pass  # not fatal

    def _heartbeat(self):
        """Refresh mtime periodically to indicate lock holder is alive."""
        while not self._stop_event.is_set():
            if os.path.exists(self.lock_path):
                try:
                    now = time.time()
                    os.utime(self.lock_path, (now, now))
                except Exception:
                    pass
            time.sleep(self.heartbeat_interval)

    def acquire(self):
        """Acquire the lock with timeout, stale cleanup, and start heartbeat."""
        self._remove_stale()

        try:
            self.lock.acquire(timeout=self.timeout)
            self._write_metadata()
            self.logger.debug("Acquired: %s", self.lock_path)
        except Timeout:
            raise Timeout(f"Timeout waiting for lock: {self.lock_path}")

        # Start heartbeat thread
        self._stop_event.clear()
        self._heartbeat_thread = threading.Thread(
            target=self._heartbeat, daemon=True
        )
        self._heartbeat_thread.start()

    def release(self):
        """Stop heartbeat and release the lock."""
        self._stop_event.set()
        if self._heartbeat_thread and self._heartbeat_thread.is_alive():
            self._heartbeat_thread.join(timeout=2)

        try:
            self.lock.release()
            self.logger.debug("Released: %s", self.lock_path)
        except Exception:
            pass

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()


"""AQUA module"""
from .version import __version__
from .graphics import plot_single_map, plot_maps, plot_single_map_diff, plot_timeseries
from .graphics import plot_hovmoller
from .graphics import plot_lat_lon_profiles, plot_seasonal_lat_lon_profiles
from .drop import Drop
from .reader import Reader, Streaming, show_catalog_content
from .regridder import Regridder
from .gridbuilder import GridBuilder
from .fldstat import FldStat
from .fixer import Fixer
from .accessor import AquaAccessor
from .histogram import histogram


__all__ = ["plot_single_map", "plot_maps", "plot_single_map_diff", "plot_timeseries",
           "plot_hovmoller", "histogram",
           "plot_lat_lon_profiles", "plot_seasonal_lat_lon_profiles",
           "Drop",
           "Reader", "Streaming", "show_catalog_content",
           "Regridder", "GridBuilder", 
           "Fixer", "FldStat"]



"""Regridding utilities."""

import os

def check_existing_file(filename):
    """
    Checks if an area/weights file exists and is valid.
    Return true if the file has some records.
    """
    return os.path.exists(filename) and os.path.getsize(filename) > 0


def validate_reader_kwargs(reader_kwargs):
    """
    Validate the reader kwargs.
    """
    if not reader_kwargs:
        raise ValueError("reader_kwargs must be provided.")
    for key in ["model", "exp", "source"]:
        if key not in reader_kwargs:
            raise ValueError(f"reader_kwargs must contain key '{key}'.")
    return reader_kwargs



"""Helpers to resolve AQUA configuration directories and files."""
import os

class ConfigLocator:
    """
    Helper to resolve AQUA configuration directories/files.
    """

    def __init__(self, filename='config-aqua.yaml', configdir=None, logger=None):
        self.filename = filename
        self._configdir = configdir
        self.logger = logger

    @property
    def configdir(self):
        if self._configdir is None:
            self._configdir = self._discover_config_dir()
        return self._configdir

    @property
    def config_file(self):
        return os.path.join(self.configdir, self.filename)

    def _discover_config_dir(self):
        """
        Search for the configuration directory in a list of predefined directories.
        """
        configdirs = []

        aquaconfigdir = os.environ.get('AQUA_CONFIG')
        if aquaconfigdir:
            configdirs.append(aquaconfigdir)

        homedir = os.environ.get('HOME')
        if homedir:
            configdirs.append(os.path.join(homedir, '.aqua'))

        for configdir in configdirs:
            if os.path.exists(os.path.join(configdir, self.filename)):
                if self.logger:
                    self.logger.debug('AQUA installation found in %s', configdir)
                return configdir

        raise FileNotFoundError(f"No config file {self.filename} found in {configdirs}")




"""Tests for the aqua.datamodel module."""
import xarray as xr
import numpy as np
import pytest
from aqua import Reader
from aqua.core.data_model import CoordTransformer, CoordIdentifier

@pytest.mark.aqua
class TestDataModel():

    @pytest.fixture
    def data(self):
        return xr.Dataset(
            {
                "temperature": (["level", "lat", "lon", "deeepth", "time"], np.random.rand(5, 3, 4, 3, 2)),
                "wind": (["height", "lat", "lon", "time"], np.random.rand(4, 3, 4, 2)),
            },
            coords={
                "level": [1000, 850, 700, 500, 300], 
                "LATITUDE": [10, 20, 30],
                "longi": [100, 110, 120, 130],
                "deeepth": [0, 10, 20],
                "timing": ["2023-01-01", "2023-01-02"],
                "height": [0, 5, 10, 15],
            },
        )
    
    def test_coords_error(self, data):
        """Error case"""

        with pytest.raises(TypeError, match="coords must be an Xarray Coordinates object."):
            CoordIdentifier(data, loglevel='debug')

        with pytest.raises(TypeError, match="data must be an Xarray Dataset or DataArray object."):
            CoordTransformer(data.coords)

        coord = CoordTransformer(data, loglevel='debug')
        with pytest.raises(TypeError, match="name must be a string."):
            coord.transform_coords(name=123)
        with pytest.raises(FileNotFoundError):
            coord.transform_coords(name="antani")
        
    def test_basic_transform_vertical(self):
        """Basic test for the CoordTransformer class."""

        reader = Reader(model="FESOM", exp="test-pi", source="original_3d", fix=False) 
        data = reader.retrieve()

        # case for multiple vertical coordinates, ignore along the vertical
        new = CoordTransformer(data, loglevel='debug').transform_coords()
        assert "nz1" in new.coords
        assert "nz" in new.coords

        # case for single vertical coordinate, convert it
        new = CoordTransformer(data['temp'], loglevel='debug').transform_coords()

        assert "depth" in new.coords
        assert "nz1" not in new.coords
        assert "idx_depth" in new.coords

    def test_basic_transform_height(self):
        """Test for height coordinate transformation."""

        reader = Reader(model="ICON", exp="test-r2b0", source="short", loglevel="warning", fix=False)
        data = reader.retrieve()
        new = CoordTransformer(data, loglevel='debug').transform_coords()
        assert "height" in new.coords

    def test_basic_transform(self):
        """Basic test for the CoordTransformer class."""

        reader = Reader(model="IFS", exp="test-tco79", source="long", fix=False)
        data = reader.retrieve(var='2t')

        new = CoordTransformer(data, loglevel='debug').transform_coords()

        assert "lon" in new.coords
        assert "X" == new['lon'].attrs["axis"]
        assert "degrees_east" == new['lon'].attrs["units"]

        assert "lat" in new.coords
        assert "Y" == new['lat'].attrs["axis"]
        assert "degrees_north" == new['lat'].attrs["units"]

    def test_bounds(self):
        """Test for bounds fixing and unit conversion."""

        data = xr.open_dataset("AQUA_tests/grids/IFS/tco79_grid.nc")
        new = CoordTransformer(data, loglevel='debug').transform_coords()

        assert "lon_bnds" in new.data_vars
        assert "lat_bnds" in new.data_vars
        assert new["lat"].max().values > 89
        assert new["lat_bnds"].max().values > 89
        assert "degrees_north" == new['lat'].attrs["units"]

    def test_fake_weird_case(self, data):
        """Test for more complex cases."""

        data["level"].attrs = {"units": "hPa"}
        data['longi'].attrs = {"units": "degrees_east"}
        data['LATITUDE'].attrs = {"units": "degrees_north"}
        data['deeepth'].attrs = {"standard_name": "depth"}
        data = data.rename({"timing": "time"})

        new = CoordTransformer(data, loglevel='debug').transform_coords()
        assert "lat" in new.coords
        assert "lon" in new.coords
        assert "level" not in new.coords
        assert "plev" in new.coords
        assert "time" in new.coords
        assert "Pa" == new["plev"].attrs["units"]
        assert new["plev"].max().values == 100000
        assert "depth" in new.coords

    def test_fake_weird_case_second(self, data):
        """Test for more complex cases."""

        data["level"].attrs = {"standard_name": "air_pressure"}
        data['timing'].attrs = {"standard_name": "time"}
        data['longi'].attrs = {"axis": "X"}
        data['LATITUDE'].attrs = {"axis": "Y"}
        data['deeepth'].attrs = {"long_name": "so much water depth"}

        new = CoordTransformer(data, loglevel='debug').transform_coords()
        assert "lat" in new.coords
        assert "lon" in new.coords
        assert "plev" in new.coords
        assert "depth" in new.coords
        assert "time" in new.coords

    def test_fake_weird_case_third(self, data):
        """Test for more complex cases."""

        data["level"].attrs = {"units": "patate"}
        data['timing'].attrs = {"axis": "T"}
        data = data.rename({"deeepth": "depth"})

        new = CoordTransformer(data, loglevel='debug').transform_coords()
        assert "time" in new.coords
        assert "depth" in new.coords

    def test_ranking_case(self, data):
        """Test for ranking functionality in CoordIdentifier."""

        data = data.rename({"LATITUDE": "lat"})
        data['longi'].attrs = {"axis": "Y"}
        
        identifier = CoordIdentifier(data.coords, loglevel='debug')
        coord_dict = identifier.identify_coords()

        # Check that only one coordinate is identified for each type
        assert coord_dict['longitude'] is None
        assert coord_dict['latitude']['name'] == 'lat'

    def same_score_ranking_case(self, data):
        """Test for conflict ranking functionality in CoordIdentifier."""

        data = data.rename({"LATITUDE": "lat"})
        data = data.rename({"longi": "latitude"})
        
        identifier = CoordIdentifier(data.coords, loglevel='debug')
        coord_dict = identifier.identify_coords()

        # No coordinate should be identified due to same score
        assert coord_dict['longitude'] is None
        assert coord_dict['latitude'] is None




import pytest
import xarray
import numpy as np
import pandas as pd

from aqua import Reader
from conftest import LOGLEVEL

loglevel = LOGLEVEL

@pytest.fixture(
    params=[
        ("IFS", "test-tco79", "short_masked", "2t")
    ]
)

def reader_arguments(request):
    return request.param

def double(x : xarray.Dataset) -> xarray.Dataset:
    return x * 2

def shift_time(x :xarray.Dataset) -> xarray.Dataset:
    """ Shift the time of 10 years """
    shifted_time = pd.to_datetime(x['time'].values) + pd.DateOffset(years=1)

    shifted_data = x.copy(deep=True)
    shifted_data['time'] = shifted_time

    return shifted_data

@pytest.mark.aqua
class TestPreproc():
    """Test different preprocessing functions"""

    def test_preproc_double(self, reader_arguments):
        """
        Test a preprocess doubling the data
        """
        model, exp, source, variable = reader_arguments

        reader = Reader(model=model, exp=exp, source=source, loglevel=loglevel)
        data = reader.retrieve()
        
        reader_preproc = Reader(model=model, exp=exp, source=source,
                                preproc=double, loglevel=loglevel)
        data_preproc = reader_preproc.retrieve()

        assert (data[variable] * 2).equals(data_preproc[variable])

    def test_preproc_shift_time(self, reader_arguments):
        """
        Test a preprocess shifting the time of 1 year
        """
        model, exp, source, _ = reader_arguments

        reader = Reader(model=model, exp=exp, source=source, loglevel=loglevel)
        data = reader.retrieve()
        
        reader_preproc = Reader(model=model, exp=exp, source=source,
                                preproc=shift_time, loglevel=loglevel)
        data_preproc = reader_preproc.retrieve()

        assert np.array_equal((pd.to_datetime(data['time'].values) + pd.DateOffset(years=1)).values,
                             data_preproc['time'].values)


from aqua.core.configurer import ConfigPath
from .util import to_list

def replace_intake_vars(path: str, catalog: str | None = None) -> str:
    """
    Replace the intake jinja vars into a string for a predefined catalog

    Args:
        path: the original path that you want to update with the intake variables
        catalog:  the catalog name where the intake vars must be read

    Returns:
        The updated path with the intake variables replaced.
    """

    # We exploit of configurerto get info on intake_vars so that we can replace them in the urlpath
    Configurer = ConfigPath(catalog=catalog)
    _, intake_vars = Configurer.get_machine_info()
    
    # loop on available intake_vars, replace them in the urlpath
    for name in intake_vars.keys():
        replacepath = intake_vars[name]
        if replacepath is not None and replacepath in path:
            # quotes used to ensure that then you can read the source
            path = path.replace(replacepath, "{{ " + name + " }}")

    return path


def replace_urlpath_jinja(block: dict, value: str, name: str, default: str | None = None) -> dict:
    """
    Replace the urlpath in the catalog entry with the given jinja parameter and
    add the parameter to the parameters block

    Args:
        block (dict): The catalog entry generated by `catalog_entry_details' to be updated
        value (str): The value to replace in the urlpath (e.g., 'r1', 'global', 'mean')
        name (str): The name of the parameter to add to the parameters block
                    and to be used in the urlpath (e.g., 'realization', 'region', 'stat')
        default(optional, str): The default value for the parameter. If the value is equal to the default,
                 no parameter is created and the urlpath is not modified.
        

    Returns:
        dict: The updated catalog entry block
    """
    if not value:
        return block
    
    # TODO: verify this is needed 
    # return if the value is the default one, no need to create a parameter
    if default is not None and value == default:
        return block

    # this loop is a bit tricky but is made to ensure that the right value is replaced
    # it works on multiple urlpath (list) or single urlpath (str)
    urlpath = to_list(block['args']['urlpath'])
    for i, _ in enumerate(urlpath):
        for character in ['_', '/', '.']:
            urlpath[i] = urlpath[i].replace(character + value + character, character + "{{" + name + "}}" + character)
    block['args']['urlpath'] = urlpath if len(urlpath) > 1 else urlpath[0]

    # add the parameter to the parameters block
    if 'parameters' not in block:
        block['parameters'] = {}
    if name not in block['parameters']:
        block['parameters'][name] = {}
        block['parameters'][name]['description'] = f"Parameter {name}"
        block['parameters'][name]['default'] = value
        block['parameters'][name]['type'] = 'str'
        block['parameters'][name]['allowed'] = [value]
    else:
        if value not in block['parameters'][name]['allowed']:
            block['parameters'][name]['allowed'].append(value)

    return block


def replace_urlpath_wildcard(block: dict, value: str) -> dict:
    """
    Replace the urlpath in the catalog entry with "*" 

    Args:
        block (dict): The catalog entry generated by `catalog_entry_details' to be updated
        value (str): The value to replace in the urlpath (e.g., 'r1', 'global', 'mean')

    Returns:
        dict: The updated catalog entry block
    """
    if not value:
        return block
    
    # this loop is a bit tricky but is made to ensure that the right value is replaced
    for character in ['_', '/', '.']:
        block['args']['urlpath'] = block['args']['urlpath'].replace(
            character + value + character, character + "*" + character)
        
    return block


from typing import Optional, Tuple
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from aqua.core.logger import log_configure
from aqua.core.util import evaluate_colorbar_limits, unit_to_latex
from .styles import ConfigStyle


def plot_vertical_profile(data: xr.DataArray, var: str= None,
                          lev_name: str = "plev", x_coord: str = "lat",
                          lev_min: Optional[float] = None,lev_max: Optional[float] = None,
                          vmin: Optional[float] = None, vmax: Optional[float] = None,
                          nlevels: int = 18, 
                          title: Optional[str] = None, title_size: Optional[int] = 16,
                          style: Optional[str] = None,
                          logscale: bool = False,
                          grid: bool = True,
                          add_land: bool = False,
                          cbar: bool = True,
                          cmap: str = "RdBu_r",
                          cbar_label: Optional[str] = None,
                          return_fig: bool = False, figsize: Tuple[int, int] = (10, 8),
                          fig: Optional[plt.Figure] = None, ax: Optional[plt.Axes] = None,
                          ax_pos: Tuple[int, int, int] = (1, 1, 1),
                          loglevel: str = "WARNING"):
    """
    Plots a zonal mean vertical profile.

    Args:
        data: DataArray to plot.
        var (str): Variable name for labeling purposes.
        lev_name (str): Name of the vertical levels (default 'plev').
        x_coord (str): Name of the horizontal coordinate (default 'lat').
        lev_min (float, optional): Minimum vertical level to plot.
        lev_max (float, optional): Maximum vertical level to plot.
        vmin (float, optional): Minimum colorbar limit.
        vmax (float, optional): Maximum colorbar limit.
        nlevels (int): Number of contour levels. Default is 18.
        title (str, optional): Plot title.
        title_size (int, optional): Title font size. Default is 16.
        style (str, optional): Plot style (default aqua style).
        logscale (bool, optional): Use log scale for y-axis if True.
        grid (bool, optional): If True, display grid lines on the plot.
        add_land (bool, optional): If True, shade land areas in the background.
        cbar (bool, optional): If True, display colorbar.
        cmap (str, optional): Colormap to use (default 'RdBu_r').
        cbar_label (str, optional): Label for the colorbar. Generated from data if None.
        return_fig (bool, optional): If True, return (fig, ax).
        figsize (Tuple[int, int], optional): Figure size.
        fig (plt.Figure, optional): Optional figure to plot on.
        ax (plt.Axes, optional): Optional axes to plot on.
        ax_pos (Tuple[int, int, int], optional): Position of subplot.
        loglevel (str, optional): Logging level.
    """
    logger = log_configure(loglevel, "plot_vertical_profile")
    ConfigStyle(style=style, loglevel=loglevel)

    # Select vertical levels
    lev_min = lev_min or data[lev_name].min().item()
    lev_max = lev_max or data[lev_name].max().item()
    mask = (data[lev_name] >= lev_min) & (data[lev_name] <= lev_max)
    data = data.sel({lev_name: data[lev_name].where(mask, drop=True)})

    # Ensure reasonable number of levels
    nlevels = max(2, int(nlevels))

    # Auto colorbar limits if not provided
    if vmin is None or vmax is None:
        vmin, vmax = float(data.min()), float(data.max())
        if vmin * vmax < 0:  # symmetric around 0
            vmax = max(abs(vmin), abs(vmax))
            vmin = -vmax

    levels = np.linspace(vmin, vmax, nlevels)

    # Prepare figure and axis
    fig = fig or plt.figure(figsize=figsize)
    ax = ax or fig.add_subplot(*ax_pos)

    # Plot
    cax = ax.contourf(data[x_coord], data[lev_name], data,
                      cmap=cmap, levels=levels, extend="both")
    if logscale:
        ax.set_yscale("log")
    ax.set_xlabel("Latitude") if x_coord == "lat" else x_coord
    ax.set_ylabel("Pressure Level (Pa)" if lev_name == "plev" else lev_name)
    ax.invert_yaxis()
    if cbar:
        if cbar_label is None:
            units = data.attrs.get('units', '')
            units_latex = unit_to_latex(units) if units else ''
            cbar_label = f"{var} [{units_latex}]" if units_latex else f"{var}"
        fig.colorbar(cax, ax=ax, label=cbar_label)
    if grid:
        ax.grid(True)

    if add_land:
        logger.debug("Adding land")
        ax.set_facecolor(color='grey')

    if title:
        logger.debug("Setting title to %s", title)
        ax.set_title(title, fontsize=title_size)

    if return_fig:
        logger.debug("Returning figure and axes")
        return fig, ax


def plot_vertical_profile_diff(data: xr.DataArray, data_ref: xr.DataArray,
                               var: str, 
                               lev_name: str = "plev", x_coord: str = "lat",
                               lev_min: Optional[float] = None, lev_max: Optional[float] = None,
                               vmin: Optional[float] = None, vmax: Optional[float] = None,
                               vmin_contour: Optional[float] = None, vmax_contour: Optional[float] = None,
                               sym_contour: bool = False, add_contour: bool = False,
                               nlevels: int = 18,
                               title: Optional[str] = None, title_size: Optional[int] = 16,
                               style: Optional[str] = None,
                               return_fig: bool = False, fig: Optional[plt.Figure] = None,
                               ax: Optional[plt.Axes] = None, ax_pos: Tuple[int, int, int] = (1, 1, 1),
                               loglevel: str = "WARNING",
                                **kwargs):
    """
    Plot the difference (data - data_ref) as vertical profile.
    Optionally add contour lines of the reference data.

    Args:
        data (xr.DataArray): Dataset to plot.
        data_ref (xr.DataArray): DataArrays to compare. A contour of this will be added if add_contour is True.
        var (str): Variable name for labeling purposes.
        lev_name (str): Name of the vertical levels. Default is 'plev'.
        x_coord (str): Name of the horizontal coordinate.
        lev_min (float, optional): Minimum vertical level to plot.
        lev_max (float, optional): Maximum vertical level to plot.
        vmin (float, optional): Minimum colorbar limit.
        vmax (float, optional): Maximum colorbar limit.
        vmin_contour (float, optional): Minimum contour limit.
        vmax_contour (float, optional): Maximum contour limit.
        sym_contour (bool, optional): If True, contour limits symmetric around zero.
        add_contour (bool, optional): If True, overlay contour lines from reference data.
        nlevels (int, optional): Number of contour levels.
        title (str, optional): Plot title.
        title_size (int, optional): Title font size. Default is 16.
        style (str, optional): Plot style (default aqua style).
        return_fig (bool, optional): If True, return (fig, ax).
        fig (plt.Figure, optional): Optional figure to plot on.
        ax (plt.Axes, optional): Optional axes to plot on.
        ax_pos (Tuple[int, int, int], optional): Position of subplot.
        loglevel (str, optional): Logging level.
        **kwargs: Additional arguments passed to plot_vertical_profile.
    """
    logger = log_configure(loglevel, "plot_vertical_profile_diff")
    ConfigStyle(style=style, loglevel=loglevel)

    # Difference
    diff = data - data_ref

    fig, ax = plot_vertical_profile(
        diff, var=var, lev_min=lev_min, lev_max=lev_max,
        vmin=vmin, vmax=vmax, nlevels=nlevels,
        style=style, return_fig=True, fig=fig, ax=ax, ax_pos=ax_pos,
        loglevel=loglevel, **kwargs)

    # Add contours of reference data
    if add_contour:
        # Contour limits
        if vmin_contour is None or vmax_contour is None:
            vmin_contour, vmax_contour = evaluate_colorbar_limits([data], sym=sym_contour)

        levels = np.linspace(vmin_contour, vmax_contour, max(2, int(nlevels)))
        data_common = data.sel({lev_name: diff[lev_name]})

        cs = ax.contour(data_common[x_coord], data_common[lev_name], data_common,
                        levels=levels, colors="k", linewidths=0.5)
        fmt = {lvl: f"{lvl:.1e}" if (abs(lvl) < 0.1 or abs(lvl) > 1000)
               else f"{lvl:.1f}" for lvl in cs.levels}
        ax.clabel(cs, fmt=fmt, fontsize=6, inline=True)

    if title:
        logger.debug("Setting title to %s", title)
        ax.set_title(title, fontsize=title_size)

    if return_fig:
        logger.debug("Returning figure and axes")
        return fig, ax



"""
DROP Output Path Builder
"""

import os
from typing import Optional
from aqua.core.util import format_realization


class OutputPathBuilder:
    """
    Class to build output paths for DROP data files.
    """

    def __init__(self, catalog: str, model: str, exp: str,
                 resolution: Optional[str] = None,
                 realization: Optional[str] = None,
                 frequency: Optional[str] = None, stat: Optional[str] = None,
                 region: Optional[str] = None, level: Optional[str] = None,
                 **kwargs):
        """
        Initialize the OutputPathBuilder with the necessary parameters.
        Params:
            catalog (str): Catalog name.
            model (str): Model name.
            exp (str): Experiment name.
            realization (str): Realization name. Default is 'r1'.
            resolution (str): Resolution. Default is None.
            frequency (str): Frequency. Default is None.
            stat (str): Statistic type. Default is 'nostat'.
            region (str): Region. Default is 'global'.
            level (str): Level. Default is None.
            kwargs: Additional keyword arguments for flexibility.
        """

        self.catalog = catalog
        self.model = model
        self.exp = exp
        self.resolution = resolution

        self.realization = format_realization(realization)  # ensure realization is formatted correctly
        self.frequency = frequency if frequency is not None else 'native'
        self.stat = stat if stat is not None else 'nostat'
        self.region = region if region is not None else 'global'

        self.level = level
        self.kwargs = kwargs or {}

    def build_path(self, basedir, var, year=None, month=None, day=None):
        """Ceate the full path to the output file.

        Args:
            basedir (str): Base directory for the output files.
            var (str): Variable name to include in the filename. Can be a wildcard.
            year (int, optional): Year to include in the filename. Defaults to None.
                                  Can be a wildcard.
            month (int, optional): Month to include in the filename. Defaults to None.
                                   Can be a wildcard.
            day (int, optional): Day to include in the filename. Defaults to None.
                                 Can be a wildcard.
        Returns:
            str: The full path to the output file.
        """
        folder = self.build_directory()
        filename = self.build_filename(var, year, month, day)
        return os.path.join(basedir, folder, filename)

    def build_directory(self):
        """
        Create the output directory based on the class parameters.

        Returns:
            str: The directory path for the output files.
        """
        parts = [
            self.catalog, self.model, self.exp, self.realization,
            self.resolution, self.frequency, self.stat, self.region
        ]
        folder = os.path.join(*[p for p in parts if p])
        return folder

    def build_filename(self, var=None, year=None, month=None, day=None):
        """
        Create the filename based on the class parameters.
        Variable and year are set as wildcards by default if not provided.
        Date format is forced to be zero-padded (e.g., 2023, 01, 01).

        Args:
            var (str, optional): Variable name to include in the filename. Defaults to None. Can be a wildcard.
            year (int, optional): Year to include in the filename. Defaults to None. Can be a wildcard.
            month (int, optional): Month to include in the filename. Defaults to None. Can be a wildcard.
            day (int, optional): Day to include in the filename. Defaults to None. Can be a wildcard.

        Returns:
            str: The filename for the output file.
        """

        # Use the provided variable or default to wildcard '*'
        var = "*" if var is None else var
        year = "*" if year is None else year

        # specific case for potential levels
        varname = f"{var}{self.level}" if self.level else var

        components = [
            varname,
            self.catalog,
            self.model,
            self.exp,
            self.realization,
            self.resolution,
            self.frequency,
            self.stat,
            self.region,
        ]

        # Convert kwargs to flat string
        kwargs_str = "_".join(f"{k}{v}" for k, v in self.kwargs.items()) if self.kwargs else ""
        components.append(kwargs_str)

        # Combine date parts with their format lengths
        date_parts = {
            'year': (year, 4),
            'month': (month, 2),
            'day': (day, 2),
        }

        # loop to format the dates
        date_components = ""
        for _, (value, length) in date_parts.items():
            date_formatted = self.format_component(value, length)
            if date_formatted:
                date_components = date_components + date_formatted

        # collapse all the component to create the final file
        filename = "_".join(str(c) for c in components + [date_components] if c) + ".nc"

        return filename

    def format_component(self, value, length):
        """
        Format a value as zero-padded string if it's not a wildcard, else return as-is.

        Args:
            value (int or str): The value to format. Can be an integer or a wildcard
            length (int): The length to zero-pad the value to, if applicable.
        """

        if value is None:
            return ""
        if self.is_wildcard(value):
            return value
        return f"{int(value):0{length}d}"

    @staticmethod
    def is_wildcard(s):
        """Guess if a string includes a wildcard"""
        return any(char in str(s) for char in ['*', '?', '[', ']'])



import operator
import re
import xarray as xr
from aqua.core.logger import log_configure, log_history

# define math operators: order is important, since defines
# which operation is done at first!
# Higher precedence operations are evaluated first
OPS = {
    '^': operator.pow,      # Power operator (highest precedence)
    '/': operator.truediv,  # Division
    "*": operator.mul,      # Multiplication  
    "-": operator.sub,      # Subtraction
    "+": operator.add       # Addition (lowest precedence)
}

class EvaluateFormula:
    """
    Class to evaluate a formula based on a string input.
    """

    def __init__(self, data: xr.Dataset, formula: str,
                 units: str = None, short_name: str = None, long_name: str = None,
                 loglevel: str = 'WARNING'):
        """
        Initialize the EvaluateFormula class.

        Args:
            data (xr.Dataset): The input data to evaluate the formula against.
            formula (str): The formula to evaluate. Supports basic arithmetic operators (+, -, *, /, ^)
                          and parentheses for grouping operations. Examples:
                          - "var1 + var2"
                          - "var1 * (var2 + var3)"
                          - "var1^2 + var2^0.5"
                          - "(var1 + var2) / (var3 - var4)"
            units (str, optional): The units of the resulting data.
            short_name (str, optional): A short name for the resulting data.
            long_name (str, optional): A long name for the resulting data.
            loglevel (str, optional): The logging level to use. Defaults to 'WARNING'.
        """
        self.loglevel = loglevel
        self.logger = log_configure(log_level=self.loglevel, log_name='EvaluateFormula')
        self.data = data
        self.formula = self.consolidate_formula(formula)
        self.units = units
        self.short_name = short_name
        self.long_name = long_name

        self.token = self._extract_tokens()

    def _evaluate(self):
        """
        Evaluate the formula using the provided data.
        Handles parentheses by recursively evaluating sub-expressions.

        Returns:
            xr.DataArray: The result of the evaluated formula as an xarray DataArray.
        """
        self.logger.debug('Evaluating formula: %s', self.formula)
        
        # Handle parentheses first by recursively evaluating sub-expressions
        formula_with_parentheses = self._handle_parentheses(self.formula)
        
        # Re-tokenize after parentheses handling
        self.token = self._extract_tokens(formula_with_parentheses)
        
        if not self.token:
            self.logger.error('No tokens extracted from the formula.')

        if len(self.token) > 1:
            if self.token[0] == '-':
                return -self.data[self.token[1]]
            return self._operations()
        return self.data[self.token[0]]
    

    def evaluate(self):
        """
        Evaluate the formula using the provided data.

        Returns:
            xr.DataArray: The result of the evaluated formula as an xarray DataArray.
        """

        out = self._evaluate()
        return self._update_attributes(out)


    def _extract_tokens(self, formula_str=None):
        """
        Tokenize the formula string into individual components.

        Returns:
            list: A list of tokens extracted from the formula.
        """
        if formula_str is None:
            formula_str = self.formula

        token = [i for i in re.split('([^\\w.]+)', formula_str) if i and i.strip()]
        return token
    
    def _handle_parentheses(self, formula_str):
        """
        Handle parentheses in the formula by recursively evaluating sub-expressions.
        
        This method finds the innermost parentheses, evaluates the expression within them,
        stores the result as a temporary variable, and replaces the parentheses expression
        with the temporary variable name.

        Args:
            formula_str (str): The formula string potentially containing parentheses.

        Returns:
            str: The formula string with parentheses resolved.
        """
        temp_var_counter = 0
        
        while '(' in formula_str:
            # Find the innermost parentheses (rightmost opening parenthesis)
            start = -1
            for i, char in enumerate(formula_str):
                if char == '(':
                    start = i
                elif char == ')' and start != -1:
                    # Found a complete parenthesis pair
                    sub_expr = formula_str[start+1:i]
                    
                    # Create temporary variable name
                    temp_var_name = f'_temp_{temp_var_counter}'
                    temp_var_counter += 1
                    
                    # Recursively evaluate the sub-expression
                    sub_evaluator = EvaluateFormula(
                        data=self.data,
                        formula=sub_expr,
                        loglevel=self.loglevel
                    )
                    sub_result = sub_evaluator._evaluate()
                    
                    # Store the result in our data dictionary for later use
                    # We'll need to modify _operations to handle temp variables
                    self.data = self.data.copy()  # Avoid modifying original data
                    self.data[temp_var_name] = sub_result
                    
                    # Replace the parentheses expression with temp variable
                    formula_str = formula_str[:start] + temp_var_name + formula_str[i+1:]
                    break
                    
        return formula_str
    
    def _operations(self):
        """
        Parsing of the operations using operator package and precedence-based evaluation.
        Now supports power operator (^) and handles temporary variables from parentheses.
        
        Operations are evaluated in order of precedence:
        1. ^ (power) - highest precedence
        2. *, / (multiplication, division)
        3. +, - (addition, subtraction) - lowest precedence

        Returns:
            xr.DataArray: The result of the evaluated formula as an xarray DataArray.
        """

        # use a dictionary to store xarray field and call them easily
        dct = {}
        for k in self.token:
            if k not in OPS:
                try:
                    dct[k] = float(k)
                except ValueError:
                    # Handle both regular variables and temporary variables from parentheses
                    if k in self.data:
                        dct[k] = self.data[k]
                    else:
                        self.logger.error(f'Variable {k} not found in data')
                        raise KeyError(f'Variable {k} not found in data')
        
        # apply operators to all occurrences, from top priority
        # Operations are processed in order of precedence (as defined in OPS dictionary)
        code = 0
        for p in OPS:
            while p in self.token:
                code += 1
                x = self.token.index(p)
                name = 'op' + str(code)
                
                # Use apply_ufunc to maintain xarray functionality
                replacer = xr.apply_ufunc(OPS.get(p), dct[self.token[x - 1]], dct[self.token[x + 1]],
                                        keep_attrs=True, dask='parallelized')
                dct[name] = replacer
                self.token[x - 1] = name
                del self.token[x:x + 2]

        return replacer
    
    def _update_attributes(self, out):
        """
        Update the attributes of the output DataArray.

        Args:
            out (xr.DataArray): The output DataArray to update.

        Returns:
            xr.DataArray: The updated DataArray with new attributes.
        """
        if self.units:
            out.attrs['units'] = self.units
        if self.short_name:
            out.attrs['short_name'] = self.short_name
            out.attrs['name'] = self.short_name
        if self.long_name:
            out.attrs['long_name'] = self.long_name
        out.attrs['AQUA_formula'] = self.formula

        msg = f'Evaluated formula: {self.formula}'
        if self.units:
            msg += f' with units: {self.units}'
        if self.short_name:
            msg += f', name: {self.short_name}'
        if self.long_name:
            msg += f', long name: {self.long_name}'

        out = log_history(out, msg)
        self.logger.debug(msg)

        return out
    
    @staticmethod
    def consolidate_formula(formula: str):
        """
        Consolidate the formula by removing unnecessary spaces and ensuring proper formatting.
        Now also validates parentheses matching.

        Args:
            formula (str): The formula to consolidate.

        Returns:
            str: The consolidated formula.
            
        Raises:
            ValueError: If parentheses are not properly matched.
        """
        # Remove spaces and ensure proper formatting
        consolidated = re.sub(r'\s+', '', formula)
        
        # Validate parentheses matching
        paren_count = 0
        for char in consolidated:
            if char == '(':
                paren_count += 1
            elif char == ')':
                paren_count -= 1
                if paren_count < 0:
                    raise ValueError("Mismatched parentheses: closing parenthesis without opening")
        
        if paren_count != 0:
            raise ValueError("Mismatched parentheses: unclosed opening parenthesis")
            
        return consolidated



import pytest
from aqua.core.graphics import ConfigStyle

@pytest.mark.graphics
class TestConfigStyle:

    def test_ConfigStyle_maptplotlib(self):
        """
        Test that ConfigStyle class is correctly initialized
        with a matplotlib style
        """
        style = 'ggplot'

        cs = ConfigStyle(style=style)
        assert cs.style == style

    def test_ConfigStyle_default(self):
        """
        Test that ConfigStyle class is correctly initialized with default style
        """
        cs = ConfigStyle()
        assert cs.style == 'aqua'

    def test_wrong_style(self):
        """
        Test that ConfigStyle class is correctly initialized with wrong style
        """
        with pytest.raises(OSError):
            ConfigStyle(style='wrong_style')
